def get_filename(filepath : str):
    file = filepath.split("/")[-1]
    return file.split(".")[0]

def get_extension(filepath : str):
    return filepath.split(".")[-1]

def replace_extension(path, new_extension):
    return ".".join(path.split(".")[:-1]) + new_extension

def keyify(*args):
    if len(args)==1:
        key = [x for x in args[0]]
    else :
        key = [x for x in args]
    key.sort()
    return tuple(key)

class Logger:
    def __init__(self, name = "Logger", verbose=True):
        self.name = name
        self.verbose = verbose

    def log(self, *messages):
        if self.verbose:
            print(*((f"[{self.name}]",) + messages))


from .base import FrameField
from ...mesh.mesh_attributes import ArrayAttribute, Attribute, Attribute
from ...mesh.datatypes import *
from ...import attributes
from ..features import FeatureEdgeDetector
from ... import geometry as geom
from ...geometry import Vec
from ...utils.maths import *
from ... import operators

from math import pi
import numpy as np
import cmath
import scipy.sparse as sp
from scipy.sparse import linalg
from osqp import OSQP

class _BaseFrameField2DVertices(FrameField): 

    @allowed_mesh_types(SurfaceMesh)
    def __init__(self, supporting_mesh : SurfaceMesh, order:int= 4, feature_edges:bool=False, verbose=True):
        super().__init__(verbose=verbose)
        self.mesh : SurfaceMesh = supporting_mesh
        self.order : int = order

        self.vbaseX : Attribute = None # local basis X vector (tangent)
        self.vbaseY : Attribute = None # local basis Y vector (tangent)
        self.vnormals : Attribute = None # local basis Z vector (normal)

        self.angles : Attribute = None # angles of every triangle corner
        self.defect : Attribute = None # sum of angles around a vertex (/!\ not 'real' defect which is 2*pi - this)
        self.cot : Attribute = None
        self.parallel_transport = dict() # (A,B) -> direction (angle) of edge (A,B) in local basis of A
        self._curvature : Attribute = None # gaussian curvature of each triangle (computed from the parallel transport)

        self.var = np.zeros(len(self.mesh.vertices), dtype=complex)

        self.features : bool = feature_edges # whether feature edges are enabled or not
        self.feat : FeatureEdgeDetector = None

    def _initialize_attributes(self):
        self.angles = attributes.corner_angles(self.mesh)
        self.cot = attributes.cotangent(self.mesh) # re-uses angle attribute
        self.vnormals = attributes.vertex_normals(self.mesh, interpolation="angle") # re-uses angle attribute
        # build defects from angles: 
        self.defect = ArrayAttribute(float, len(self.mesh.vertices))
        for iC, C in enumerate(self.mesh.face_corners):
            self.defect[C] = self.defect[C] + self.angles[iC]

    def _initialize_features(self):
        self.feat = FeatureEdgeDetector(only_border = not self.features, verbose=self.verbose)(self.mesh)

    def _initialize_basis(self):
        """Init self.vbaseX, self.vbaseY and self.edge_angles"""
        n_vert = len(self.mesh.vertices)
        self.vbaseX = ArrayAttribute(float, n_vert, 3)
        self.vbaseY = ArrayAttribute(float, n_vert, 3)

        for u in range(n_vert):
            # find basis vector -> first edge
            P = Vec(self.mesh.vertices[u])
            N = self.vnormals[u]
            # extract basis edge
            v = self.mesh.connectivity.vertex_to_vertex(u)[0]
            E = self.mesh.vertices[v] - P
            X = Vec.normalized(E - np.dot(E,N)*N) # project on tangent plane
            self.vbaseX[u] = X
            self.vbaseY[u] = geom.cross(N,X)

            # initialize angles of every edge in this basis
            ang = 0.
            if self.mesh.is_vertex_on_border(u):
                # for v in self.mesh.connectivity.vertex_to_vertex(u):
                #     T,iu,_ = self.mesh.half_edges.adj(u,v)
                #     self.parallel_transport[(u,v)] = ang
                #     ang += self.angles[(T,iu)]

                fst, lst = self.mesh.connectivity.vertex_to_vertex(u)[0], self.mesh.connectivity.vertex_to_vertex(u)[-1]
                pfst, plst = (self.mesh.vertices[x] for x in (fst, lst))
                comp_angle = geom.signed_angle_3pts(plst,P,pfst, N) # complementary angle, ie "exterior" angle between two edges on the boundary
                comp_angle = 2*pi + comp_angle if comp_angle<0 else comp_angle
                
                for v in self.mesh.connectivity.vertex_to_vertex(u):
                    T = self.mesh.half_edges.adj(u,v)[0]
                    self.parallel_transport[(u,v)] = ang * 2 * pi / (self.defect[u] + comp_angle)
                    #self.parallel_transport[(u,v)] = ang * pi / self.defect[u]
                    if T is None : continue
                    c = self.mesh.connectivity.vertex_to_corner_in_face(u,T)
                    ang += self.angles[c]
            else:
                for v in self.mesh.connectivity.vertex_to_vertex(u):
                    T = self.mesh.half_edges.adj(u,v)[0]
                    c = self.mesh.connectivity.vertex_to_corner_in_face(u,T)
                    self.parallel_transport[(u,v)] = ang * 2 * pi / self.defect[u]
                    ang += self.angles[c]

    def _initialize_variables(self, mean_normals=True):
        """Init self.var for feature vertices
        
        mean_normals : whether to initialize the frame field as a mean of adjacent feature edges (True), or following one of the edges (False)
        """
        if mean_normals:
            for e in self.feat.feature_edges:
                A,B = self.mesh.edges[e]
                edge = self.mesh.vertices[B] - self.mesh.vertices[A]
                vx,vy = np.dot(edge, self.vbaseX[B]), np.dot(edge, self.vbaseY[B])
                v = complex(vx, vy)
                vpow = (v/abs(v)) ** self.order
                if abs(self.var[B] + vpow)>1e-10:
                    self.var[B] += vpow

                vx,vy = np.dot(edge, self.vbaseX[A]), np.dot(edge, self.vbaseY[A])
                v = complex(vx, vy)
                vpow = (v/abs(v)) ** self.order
                if abs(self.var[A] + vpow)>1e-10:
                    self.var[A] += vpow
        else:
            for e in self.feat.feature_edges:
                A,B = self.mesh.edges[e]
                self.var[A] += cmath.rect(1,self.parallel_transport[(A,B)])**self.order
                self.var[B] += cmath.rect(1,self.parallel_transport[(B,A)])**self.order

        for A in self.feat.feature_vertices:
            self.var[A] /= abs(self.var[A])

    def _compute_attach_weight(self, A, fail_value=1e-3):
        return fail_value
        # A is area weight matrix
        lap_nopt = operators.laplacian(self.mesh, parallel_transport=None)
        try:
            eigs = sp.linalg.eigsh(lap_nopt, k=2, M=A, which="SM", tol=1e-3, maxiter=500, return_eigenvectors=False)
        except Exception as e:
            try:
                self.log("Estimation of alpha failed: {}".format(e))
                eigs = sp.linalg.eigsh(lap_nopt-0.01*sp.identity(lap_nopt.shape[0]), k=2, which="SM", tol=1e-3, maxiter=500, return_eigenvectors=False)
            except:
                self.log("Second estimation of alpha failed: taking alpha = ", fail_value)
                return fail_value
        eigs_non_zero = [e for e in eigs if abs(e)>1e-6]
        if len(eigs_non_zero)==0:
            return fail_value
        return abs(min(eigs_non_zero))

    @property
    def curvature(self):
        """Angle defect of each triangles of the mesh, computed from the parallel transport

        Returns:
            Attribute: an attribute on faces
        """
        if self._curvature is None:
            # compute curvature on the mesh
            if self.mesh.faces.has_attribute("curvature"):
                self._curvature = self.mesh.faces.get_attribute("curvature")
            else:
                self._curvature = self.mesh.faces.create_attribute("curvature", float)
                for iF, (A,B,C) in enumerate(self.mesh.faces):
                    v = 1+0j
                    for a,b in [(A,B) ,(B,C), (C,A)]:
                        v *= cmath.rect(1., self.parallel_transport[(b,a)] - self.parallel_transport[(a,b)] - pi)
                    self._curvature[iF] = cmath.phase(v)
        return self._curvature

    def flag_singularities(self):
        """Compute singularity data.

        Creates 3 attributes:
            - an attribute "curvature" on faces storing the gaussian curvature (used as a corrective term for computing the values of the singularities)
            - an attribute "singuls" on faces storing the value (+- 1) of singularities (eventually 0 for non singular triangles)
            - an attribute "angles" on edges storing the angle of the edge, given as the difference between the two frames
        """
        self._check_init()
        ZERO_THRESHOLD = 1e-4

        edge_rot = dict() # the rotation induced by the frame field on every edge
        if self.mesh.edges.has_attribute("angles"):
            edge_rot_attr = self.mesh.edges.get_attribute("angles")
            edge_rot_attr.clear()
        else:
            edge_rot_attr = self.mesh.edges.create_attribute("angles", float, 1)
        for ie,(A,B) in enumerate(self.mesh.edges):
            fA,fB = self.var[A], self.var[B] # representation complex for A and B
            aA,aB = self.parallel_transport[(A,B)], self.parallel_transport[(B,A)] # local basis orientation for A and B
            uB = roots(fB, self.order)[0]
            abs_angles = [abs( angle_diff( cmath.phase(uB) - aB - pi, cmath.phase(uA)-aA)) for uA in roots(fA, self.order)]
            angles = [angle_diff( cmath.phase(uB) - aB - pi, cmath.phase(uA)-aA) for uA in roots(fA, self.order)]
            i_angle = np.argmin(abs_angles)
            edge_rot[(A,B)] = angles[i_angle]
            edge_rot[(B,A)] = -angles[i_angle]
            edge_rot_attr[ie] = -angles[i_angle]

        if self.mesh.faces.has_attribute("singuls"):
            singuls = self.mesh.faces.get_attribute("singuls")
        else:
            singuls = self.mesh.faces.create_attribute("singuls", int)
        for id_face,(A,B,C) in enumerate(self.mesh.faces):
            angle = 0
            for u,v in [(A,B), (B,C), (C,A)]:
                angle += edge_rot[(u,v)]
            angle += self.curvature[id_face]
            if angle>ZERO_THRESHOLD:
                singuls[id_face] = 1
            elif angle<-ZERO_THRESHOLD:
                singuls[id_face] = -1

    def export_as_mesh(self, repr_vector=False) -> PolyLine:
        """
        Exports the frame field as a mesh for visualization.

        Returns:
            PolyLine: the frame field as a mesh object
        """
        self._check_init()
        FFMesh = PolyLine()
        L = attributes.mean_edge_length(self.mesh)/3
        n = self.order + 1
        for id_vertex, P in enumerate(self.mesh.vertices):
            E,N = self.vbaseX[id_vertex], self.vnormals[id_vertex]
            if repr_vector:
                # Representation vector only: 
                angle = cmath.phase(self.var[id_vertex])
                r = geom.rotate_around_axis(E, N, angle)
                p = P + L*r
                FFMesh.vertices += [P, p]
                FFMesh.edges.append((2*id_vertex, 2*id_vertex+1))
            else:
                # Representation of the whole frame
                angle = cmath.phase(self.var[id_vertex])/self.order
                cmplx = [geom.rotate_around_axis(E, N, angle + 2*k*pi/self.order) for k in range(self.order)]
                pts = [P + abs(self.var[id_vertex])*L*r for r in cmplx]
                FFMesh.vertices.append(P)
                FFMesh.vertices += pts
                FFMesh.edges += [(n*id_vertex, n*id_vertex+k) for k in range(1,n)]
        return FFMesh

class FrameField2DVertices(_BaseFrameField2DVertices):
    
    @allowed_mesh_types(SurfaceMesh)
    def __init__(self, supporting_mesh : SurfaceMesh, order : int = 4, feature_edges : bool = False, verbose=True):
        super().__init__(supporting_mesh, order, feature_edges, verbose)
                
    def initialize(self):
        self._initialize_attributes()
        self._initialize_features() # /!\ before initialize basis
        self._initialize_basis()
        self._initialize_variables()
        self.initialized = True

    def optimize(self, n_renorm=10):
        self._check_init()
        self.log("Build laplacian operator")
        lap = operators.laplacian(self.mesh, parallel_transport=self.parallel_transport, order=self.order)
        A = operators.area_weight_matrix(self.mesh).tocsc()

        if len(self.feat.feature_vertices)>0: # We have a border / feature elements -> linear solve
            self.log("Feature element detected (border and/or feature edges)")
            # Compute attach weight as smallest eigenvalue of the laplacian
            self.log("Compute attach weight")
            alpha = self._compute_attach_weight(A)
            self.log("Attach weight: {}".format(alpha))

            A = A.astype(complex)

            # Build fixed/free vertex partition
            fixedInds,freeInds = [],[]
            for v in self.mesh.id_vertices:
                if v in self.feat.feature_vertices:
                    fixedInds.append(v)
                else:
                    freeInds.append(v)
            
            lapI = lap[freeInds,:][:,freeInds]
            lapB = lap[freeInds,:][:,fixedInds]
            AI = A[freeInds, :][:,freeInds]
            # for lapI and lapB, only lines of freeInds are relevant : lines of fixedInds link fixed variables -> useless constraints
            valB = lapB.dot(self.var[fixedInds]) # right hand side

            self.log("Initial solve of linear system")
            res = linalg.spsolve(lapI, -valB) # first system solved without diffusion
            self.var[freeInds] = res

            if n_renorm>0:
                self.log(f"Solve linear system {n_renorm} times with diffusion")
                mat = lapI - alpha * AI
                for _ in range(n_renorm):
                    self.normalize()
                    valI2 = alpha * AI.dot(self.var[freeInds])
                    res = linalg.spsolve(mat, - valB - valI2)
                    self.var[freeInds] = res
            self.normalize()

        else: # No border -> eigensolve
            self.log("No border detected")
            self.log("Initial solve of linear system using an eigensolver")
            A = A.astype(complex)
            try:
                egv, U = sp.linalg.eigsh(lap, k=2, M=A, which="SM", tol=1e-3, maxiter=1000)
            except linalg.ArpackNoConvergence as e:
                self.log("Initial eigensolve failed :", e)
                self.log("Retry using connectivity laplacian")
                lap = operators.laplacian(self.mesh, cotan=False, parallel_transport=self.parallel_transport, order=self.order)
                egv, U = sp.linalg.eigsh(lap-0.1*sp.identity(lap.shape[0]), k=2, M=A, which="SM", tol=1e-3)
            iu = np.argmin(abs(egv))
            alpha = abs(egv[iu])
            self.log("Eigenvalues:", egv)
            self.var = U[:,iu]

            if n_renorm>0:
                self.log(f"Solve linear system {n_renorm} times with diffusion")
                mat = lap  - alpha * A
                for _ in range(n_renorm):
                    self.normalize()
                    valI2 = alpha * A.dot(self.var)
                    self.var = linalg.spsolve(mat, - valI2)
            self.normalize()

class CadFF2DVertices(FrameField2DVertices):

    @allowed_mesh_types(SurfaceMesh)
    def __init__(self, supporting_mesh : SurfaceMesh, order:int = 4, verbose=True):
        super().__init__(supporting_mesh, order, feature_edges=True, verbose=verbose)
        self.target_w : dict = None

    def _initialize_target_w(self):
        self.target_w = dict()
        for e in self.feat.feature_edges:
            A,B = self.mesh.edges[e]
            aA,aB = self.parallel_transport[(A,B)], self.parallel_transport[(B,A)] # local basis orientation for A and B
            fA = self.var[A] # representation complex for frame field at A
            fB = self.var[B] # representation complex for frame field at B
            uB = roots(fB, self.order)[0]
            abs_angles = [abs(angle_diff( cmath.phase(uB) - aB - pi, cmath.phase(uA)-aA)) for uA in roots(fA, self.order)]
            angles = [angle_diff( cmath.phase(uB) - aB - pi, cmath.phase(uA)-aA) for uA in roots(fA, self.order)]
            i_angle = np.argmin(abs_angles)
            self.target_w[e] = angles[i_angle]

        ## build start w for rotation penalty energy
        cstrfaces = attributes.faces_near_border(self.mesh, 3)
        # 1) constraints
        ncstr = len(self.feat.feature_edges)
        nvar = len(self.mesh.edges)
        Cstr = sp.lil_matrix((ncstr, nvar))
        Rhs = np.zeros(ncstr)
        for i,e in enumerate(self.feat.feature_edges):
            Cstr[i,e] = 1
            Rhs[i] = self.target_w[e]

        # objective
        D1 = sp.lil_matrix((len(self.mesh.faces), nvar))
        b1 = np.zeros(len(self.mesh.faces))
        for iT,T in enumerate(self.mesh.faces):
            A,B,C = T
            for u,v in [(A,B), (B,C), (C,A)]:
                e = self.mesh.connectivity.edge_id(u,v)
                D1[iT, e]= 1 if u<v else -1
            b1[iT] = -self.curvature[iT]
        
        D2 = sp.lil_matrix((len(cstrfaces), nvar))
        b2 = np.zeros(len(cstrfaces))
        for i,f in enumerate(cstrfaces):
            A,B,C = self.mesh.faces[f]
            for u,v in [(A,B), (B,C), (C,A)]:
                e = self.mesh.connectivity.edge_id(u,v)
                D2[i,e] = 1e3 if (u<v) else -1e3
            b2[i] = -1e3*self.curvature[f]

        I = sp.identity(len(self.mesh.edges), format="csc")
        b3 = np.zeros(len(self.mesh.edges))

        P = sp.vstack((D1,D2,I))
        b = np.concatenate([b1,b2,b3])
        b = P.transpose().dot(b)
        P = P.transpose().dot(P)
        osqp_instance = OSQP()
        osqp_instance.setup(P, b, A=Cstr.tocsc(), l=Rhs, u=Rhs, verbose=False)
        res = osqp_instance.solve().x
        for e in self.mesh.id_edges:
            self.target_w[e] = res[e]

    def initialize(self):
        self._initialize_attributes()
        self._initialize_features() # /!\ before initialize basis
        self._initialize_basis()
        self._initialize_variables()
        self._initialize_target_w()
        self.initialized = True

    def _cotan_laplacian_parallel_transport(self):
        """For CadFF, the cotan laplacian operator is modified by the target angles computed at the initialization

        Returns:
            scipy.sparse.lil_matrix: cotan laplacian as sparse matrix
        """
        n = len(self.mesh.vertices)
        mat = sp.lil_matrix((n,n), dtype=complex)
        for it, (p,q,r) in enumerate(self.mesh.faces):
            a,b,c = (self.cot[(it, k)]/2 for k in range(3))
            for (i, j, v) in [(p, q, c), (q, r, a), (r, p, b)]:
                ai, aj = self.parallel_transport[(i,j)], self.parallel_transport[(j,i)]
                e = self.mesh.connectivity.edge_id(i,j)
                w = self.target_w[e] if i<j else -self.target_w[e]
                mat[i,i] -= v
                mat[j,j] -= v
                mat[i,j] += v * cmath.rect(1., self.order*(ai - aj + w))
                mat[j,i] += v * cmath.rect(1., self.order*(aj - ai - w))
        return mat

    def _adj_laplacian_parallel_transport(self):
        """Laplacian with classic +- 1 weights 
        """
        n = len(self.mesh.vertices)
        mat = sp.lil_matrix((n,n), dtype=complex)
        for (p,q,r) in self.mesh.faces:
            for (i, j) in [(p, q), (q, r), (r, p)]:
                ai, aj = self.parallel_transport[(i,j)], self.parallel_transport[(j,i)]
                e = self.mesh.connectivity.edge_id(i,j)
                w = self.target_w[e] if i<j else -self.target_w[e]
                mat[i,i] -= 1
                mat[j,j] -= 1
                mat[i,j] += cmath.rect(1., self.order*(ai - aj + w))
                mat[j,i] += cmath.rect(1., self.order*(aj - ai - w))
        return mat

class CurvatureVertices(_BaseFrameField2DVertices):

    @allowed_mesh_types(SurfaceMesh)
    def __init__(self, mesh : SurfaceMesh, feature_edges:bool = False, verbose=True):
        super().__init__(mesh, 
            4, # order is always 4 
            feature_edges,
            verbose
        )
        self.face_areas : Attribute = None
        self.curv_mat_vert : np.ndarray = None # shape (|V|,3,3)

    def run(self, n_smooth=0):
        self.log("Compute curvature matrices")
        self.initialize()
        self.log("Optimize")
        self.optimize(n_smooth)
        self.log("Done.")
        return self

    def _initialize_attributes(self):
        super()._initialize_attributes()
        self.face_areas = attributes.face_area(self.mesh)

    def _initialize_curv_matrices(self):
        curv_mat_edges = attributes.curvature_matrices(self.mesh)
        self.curv_mat_vert = np.zeros((len(self.mesh.vertices),3,3))
        for v in self.mesh.id_vertices:
            #total_area = 0
            for e in self.mesh.connectivity.vertex_to_edge(v):
                v2 = self.mesh.connectivity.other_edge_end(e,v)
                area_ab = sum([self.face_areas[_T] for _T in self.mesh.half_edges.edge_to_triangles(v,v2) if _T is not None])
                #total_area += area_ab
                self.curv_mat_vert[v,:,:] += area_ab * curv_mat_edges[e,:,:]
            #self.curv_mat_vert[v,:,:] /= total_area

    def initialize(self):
        self._initialize_attributes()
        self._initialize_features() # /!\ before initialize basis
        self._initialize_basis()
        self._initialize_curv_matrices()
        self._initialize_variables()
        self.initialized = True

    def optimize(self, n_smooth: int = 1, normalize:bool = True):
        for v in self.mesh.id_vertices:
            if v in self.feat.feature_vertices: continue # do not override frames on boundary
            
            U,S,V = np.linalg.svd(self.curv_mat_vert[v,:,:], hermitian=True)
            # three vectors -> normal and two principal components
            # eigenvalue of normal is 0
            # PC are orthogonal (eigenvects of symmetric matrix) -> we rely on the eigenvect of greatest eigenvalue and take orthog direction

            X,Y = self.vbaseX[v], self.vbaseY[v]
            if S[0]<1e-8 : # zero matrix -> no curvature information
                self.var[v] = 0 + 0j
            else:
                # representation vector
                eig = V[0,:]
                c = complex(X.dot(eig), Y.dot(eig))
                self.var[v] = (c/abs(c))**4

        if n_smooth > 0:
            # diffuse the curvature results to get a smoother results (especially where curvature was not defined)
            lap = operators.laplacian(self.mesh, parallel_transport=self.parallel_transport, order=4)
            A = operators.area_weight_matrix(self.mesh)
            alpha = 0.1*attributes.mean_edge_length(self.mesh)
            self.log("Attach weight:", alpha)

            if len(self.feat.feature_vertices)>0:
                # Build fixed/free vertex partition
                fixedInds,freeInds = [],[]
                for v in self.mesh.id_vertices:
                    if v in self.feat.feature_vertices:
                        fixedInds.append(v)
                    else:
                        freeInds.append(v)
                lapI = lap[freeInds,:][:,freeInds]
                lapB = lap[freeInds,:][:,fixedInds]
                AI = A[freeInds, :][:,freeInds]
                # for lapI and lapB, only lines of freeInds are relevant : lines of fixedInds link fixed variables -> useless constraints
                valB = lapB.dot(self.var[fixedInds]) # right hand side
                mat = lapI - alpha * AI
                for _ in range(n_smooth):
                    valI2 = alpha * AI.dot(self.var[freeInds])
                    res = linalg.spsolve(mat, - valB - valI2)
                    self.var[freeInds] = res
                    if normalize: self.normalize()
            else:
                mat = lap  - alpha * A
                for _ in range(n_smooth):
                    valI2 = alpha * A.dot(self.var)
                    self.var = linalg.spsolve(mat, - valI2)
                    if normalize: self.normalize()



import mouette as M

### Point Clouds ###

def pc_sample1():
    m = M.mesh.new_point_cloud()
    m.vertices += [
        M.Vec(0., 0., 0.),
        M.Vec(1., 0., 0.),
        M.Vec(0., 1., 0.),
        M.Vec(0., 0., 1.),
    ]

def pc_random10():
    m = M.mesh.new_point_cloud()
    m.vertices += [
        M.Vec.random(3) for _ in range(10)
    ]

def pc_grid():
    m = M.mesh.new_point_cloud()
    for x in range(5):
        for y in range(5):
            for z in range(5):
                m.vertices.append(M.Vec(float(x), float(y), float(z)))
    return m

point_clouds = [
    pc_sample1(),
    pc_random10(),
    pc_grid()
]


import cmath
from math import pi

def roots(c : complex, pow: int, normalize=True):
    """Given a complex number c, compute and returns the four ^(1/4) roots 
    Args:
        c (complex): input complex number
        pow (int): the power of the root. Returns power-th roots of c
        normalize (bool, optional): If True, roots will have module 1. Defaults to True

    Returns:
        list of complex roots
    """
    r,t = cmath.polar(c)
    r = 1 if normalize else r**(1/pow)
    return [cmath.rect(r, (t + 2*k*pi)/pow) for k in range(pow)]

def angle_diff(a,b):
    return (a - b + pi) % (2*pi) - pi


"""
M.processing
submodule for algorithms applied on meshes
"""

from .misc import *
from .combinatorics.subdivide import *
from .combinatorics.cutting import SingularityCutter
from .features import FeatureEdgeDetector

from .paths import shortest_path, shortest_path_to_border, shortest_path_to_vertex_set
from .border import extract_border_cycle, extract_border_cycle_all, extract_curve_boundary, extract_surface_boundary

from . import trees
from .parametrization.lscm import LSCM
from .parametrization.distortion import SurfaceDistortion


import mouette as M
import pytest
from math import cos, sin, pi
import os

### Surfaces ###

def surf_triangle():
    m = M.mesh.new_surface()
    m.vertices += [
        M.Vec(0., 0., 0.),
        M.Vec(1., 0., 0.),
        M.Vec(0., 0., 1.),
    ]
    m.edges += [(0,1), (0,2), (1,2)]
    m.faces.append((0,1,2))
    m.face_corners += [0,1,2]
    return m

def surf_square():
    m = M.mesh.new_surface()
    m.vertices += [
        M.Vec(1.,0.,0.),
        M.Vec(0.,1.,0.),
        M.Vec(1.,1.,0.),
        M.Vec(1.,0.,0.),
    ]
    m.edges += [(0,1), (1,2), (2,3), (0,3), (1,3)]
    m.faces += [(0,1,3), (1,2,3)]
    m.face_corners += [0,1,2,3,1,2,3]
    return m

def surf_cube():
    m = M.mesh.new_surface()
    m.vertices += [
        M.Vec(0., 0., 0.),
        M.Vec(1., 0., 0.),
        M.Vec(1., 1., 0.),
        M.Vec(0., 1., 0.),
        M.Vec(0., 0., 1.),
        M.Vec(1., 0., 1.),
        M.Vec(1., 1., 1.),
        M.Vec(0., 1., 1.),
    ]
    m.faces += [
        
    ]
    return m

def surf_tetrahedron():
    m = M.mesh.new_surface()
    m.vertices += [
        M.Vec(0., 0., 0.),
        M.Vec(1., 0., 0.),
        M.Vec(0., 0., 1.),
        M.Vec(0., 1., 0.),
    ]
    m.edges += [(0,1), (1,2), (0,2), (0,3), (1,3), (2,3)]
    m.faces += [[0,1,2], [0,1,3], [1,2,3], [2,0,3]]
    m.face_corners += [0,1,2,0,1,3,1,2,3,2,0,3]
    return m

def surf_one_ring():
    return M.procedural.ring(10,0.1)

def surf_circle():
    return M.mesh.load("tests/data/circle.mesh")

def surf_half_sphere():
    return M.mesh.load("tests/data/half_sphere1.obj")

def surf_spline():
    return M.mesh.load("tests/data/spline03.mesh")

def surf_feat():
    return M.mesh.load("tests/data/feature2.obj")

def surf_two_pieces():
    return M.mesh.load("tests/data/two_pieces.obj")

def surf_pointy():
    return M.mesh.load("tests/data/pointy.obj")

surfaces = [
    surf_triangle(),
    surf_one_ring(),
    surf_circle(),
    surf_half_sphere(),
    surf_spline(),
    surf_feat(),
    surf_two_pieces(),
    surf_pointy()
]


import mouette as M
from mouette import geometry as geom
from math import *
import numpy as np

def test_sign():
    assert geom.sign(1.)==1
    assert geom.sign(-1.)==-1
    assert geom.sign(2.45)==1
    assert geom.sign(-1e8)==-1
    assert geom.sign(0.)==0

def test_norm():
    x = M.Vec(1.,1.,1.)
    assert geom.norm(x) == x.norm()
    assert abs(geom.norm(x) - sqrt(3))<1e-15
    x = M.Vec(3.,4.)
    assert geom.norm(x) == 5.
    a = np.array([[1,0],[0,1]])
    assert abs(geom.norm(a) - sqrt(2)) < 1e-15


def test_dot():
    x = M.Vec(1.,1.,1.)
    y = M.Vec(0.,1.,0.)
    assert geom.dot(x,y) == 1.
    y = M.Vec(-1,2,1)
    assert geom.dot(x,y) == 2.

def test_distance():
    x = M.Vec(1.,1.,1.)
    y = M.Vec(1.,0.,1.)
    assert geom.distance(x,y) == 1.

def test_cross():
    x = M.Vec(1.,0.,0.)
    y = M.Vec(0.,1.,0.)
    assert (geom.cross(x,y) == M.Vec(0.,0.,1.)).all()

def test_cotan():
    A = M.Vec(1.,0.,0.)
    B = M.Vec(0.,0.,0.)
    C = M.Vec(0.,1.,0.)
    assert geom.cotan(A,B,C)==0.

def test_angle3pts():
    A = M.Vec.random(3)
    B = M.Vec.random(3)
    C = M.Vec.random(3)
    ang = geom.angle_3pts(A,B,C)
    assert True

def test_signed_angle_2vec3D():
    A = M.Vec.random(3)
    B = M.Vec.random(3)
    N = M.Vec.random(3)
    ang = geom.signed_angle_2vec3D(A,B,N)
    assert True

def test_signed_angle_3pts():
    A = M.Vec.random(3)
    B = M.Vec.random(3)
    N = M.Vec.random(3)
    ang = geom.signed_angle_2vec3D(A,B,N)
    assert True

def test_angle_2vec2d():
    A = M.Vec(1.,0.)
    B = M.Vec(1.,1.)
    assert geom.angle_2vec2D(A,B) == pi/4

def test_angle_2vec3d():
    A = M.Vec(1.,0.,0.)
    B = M.Vec(0.,1.,0.)
    assert geom.angle_2vec3D(A,B) == pi/2

def test_face_basis():
    A = M.Vec.random(3)
    B = M.Vec.random(3)
    C = M.Vec.random(3)
    X,Y,Z = geom.face_basis(A,B,C)
    X,Y,Z = geom.face_basis((A,B,C))
    assert True

def test_triangle_area():
    A = M.Vec.random(3)
    B = M.Vec.random(3)
    C = M.Vec.random(3)
    assert geom.triangle_area(A,B,C)>=0

    A = M.Vec.random(2)
    B = M.Vec.random(2)
    C = M.Vec.random(2)
    assert geom.triangle_area_2D(A,B,C)>=0

def test_quad_area():
    A = M.Vec(0.,0.,0.)
    B = M.Vec(1.,0.,0.)
    C = M.Vec(1.,1.,0.)
    D = M.Vec(0.,1.,0.)
    assert geom.quad_area(A,B,C,D)==1.

def test_det_2x2():
    A = M.Vec(1.,0.)
    B = M.Vec(0.,1.)
    assert geom.det_2x2(A,B)==1.

def test_det_3x3():
    A = M.Vec(1.,0.,0.)
    B = M.Vec(0.,1.,0.)
    C = M.Vec(0.,0.,1.)
    assert geom.det_3x3(A,B,C)==1.
    assert geom.det_3x3(A,C,B)==-1.

def test_intersect_2lines2D():
    x = geom.intersect_2lines2D(M.Vec.zeros(3), M.Vec(1.,0.,0.), M.Vec.zeros(3), M.Vec(1.,1.,0.))
    assert geom.norm(x)<1e-15
    dir = M.Vec.random(3)
    assert geom.intersect_2lines2D(M.Vec(1.,0.,0.), dir, M.Vec(-1.,0.,0.), dir) is None

def test_circumcenter():
    A = M.Vec(-1,0.,0.)
    B = M.Vec(1.,0.,0.)
    alpha = 2*pi*np.random.random()
    C = M.Vec(cos(alpha), sin(alpha), 0.)
    assert geom.norm(geom.circumcenter(A,B,C))<1e-10


import heapq as hq
from typing import Any
from dataclasses import dataclass, field

@dataclass
class PriorityItem:
    x : Any = field(compare=False)
    priority: float

    def __lt__(self, other):
        return self.priority < other.priority

class PriorityQueue:
    def __init__(self):
        self.data = []

    @property
    def front(self) -> PriorityItem:
        return self.data[0]

    def empty(self) -> bool:
        return len(self.data) == 0

    def get(self) -> PriorityItem:
        return hq.heappop(self.data)

    def pop(self) -> PriorityItem:
        return self.get()
    
    def push(self, x, w) -> None:
        item = PriorityItem(x, w)
        hq.heappush(self.data, item)



from .sphere import sphere_fibonacci, sphere_uv, icosphere, spherify_vertices
from .cylinder import cylinder, cylindrify_edges
from .hexahedra import hexahedron_4pts, axis_aligned_cube
from .platonic import tetrahedron, hexahedron, icosahedron
from .rings import ring, flat_ring


import pytest
import mouette as M
from utils import *

@pytest.fixture
def test_container():
    container = M.mesh.mesh_data.DataContainer(id="test")
    container.append([0,1,2])
    container.append([4,5,6])
    container.append([0,3,4])
    return container

def test_attribute_sparse(test_container):
    attr = test_container.create_attribute("test_attr",float,1)
    attr[0] = 42.
    assert attr[0] == 42.
    assert attr[1] == 0.

def test_attribute_dense(test_container):
    attr = test_container.create_attribute("test_attr",float,1, dense=True)
    attr[0] = 42.
    assert attr[0] == 42.
    assert attr[1] == 0.
    assert attr[2] == 0.

def test_attribute_as_array(test_container):
    attr = test_container.create_attribute("test_attr",float,1)
    attr[0] = 42.
    attr[1] = 76.52
    attr[2] = -26.5
    diff = abs(attr.as_array() - np.array([42., 76.52, -26.5]))
    assert np.all(diff < 1e-8)

def test_attribute_sparse_cast(test_container):
    attr = test_container.create_attribute("test_attr",float,1)
    attr[0] = True
    assert attr[0] == 1.
    attr[0] = 1
    assert attr[0] == 1.

    try:
        attr[0] = "toto"
        assert False
    except:
        assert True
    
    try:
        attr[0] = 1+2j
        assert False
    except:
        assert True

def test_attribute_has_get_delete(test_container):
    assert not test_container.has_attribute("test_attr")
    test_container.create_attribute("test_attr",float,1)
    assert test_container.has_attribute("test_attr")
    attr = test_container("test_attr")
    attr.clear()
    test_container.delete_attribute("test_attr")
    assert not test_container.has_attribute("test_attr")

def test_attribute_default_value(test_container):

    attr = test_container.create_attribute("test_attr",float,1, default_value=42.)
    attr[1] = 76.5
    assert attr[0] == 42.
    assert attr[1] == 76.5
    assert attr[2] == 42.

def test_attribute_types(test_container):
    attr_bool    = test_container.create_attribute("test_bool",bool,1)
    attr_int     = test_container.create_attribute("test_int",int,1)
    attr_float   = test_container.create_attribute("test_float",float,1)
    attr_complex = test_container.create_attribute("test_complex",complex,1)
    attr_str     = test_container.create_attribute("test_str",str,1)

    attr_bool[0] = True
    assert attr_bool[0] == True

    attr_int[0] = 42
    assert attr_int[0] == 42
    
    attr_float[0] = 23.52
    assert attr_float[0] == 23.52

    attr_complex[0] = 10+22.5j
    assert attr_complex[0] == 10+22.5j 

    attr_str[0] = "bonjour"
    assert attr_str[0] == "bonjour"


def test_attribute_types_vec(test_container):
    attr_bool    = test_container.create_attribute("test_bool",bool,2)
    attr_int     = test_container.create_attribute("test_int",int,2)
    attr_float   = test_container.create_attribute("test_float",float,2)
    attr_complex = test_container.create_attribute("test_complex",complex,2)
    attr_str     = test_container.create_attribute("test_str",str,2)

    attr_bool[0] = [True, False]
    assert (attr_bool[0] == [True, False]).all()

    attr_int[0] = [42, 10]
    assert (attr_int[0] == [42,10]).all()
    
    attr_float[0] = [23.52, 35.12]
    assert (attr_float[0] == [23.52, 35.12]).all()
    attr_float[0][0] = 11.111
    assert (attr_float[0] == [11.111, 35.12]).all()

    attr_complex[0] = [10+22.5j, 9+9j]
    assert (attr_complex[0] == [10+22.5j, 9+9j] ).all()

    attr_str[0] = ["bonjour", "merci"]
    assert (attr_str[0] == ["bonjour", "merci"]).all()

def test_attribute_out_of_bounds(test_container):
    attr = test_container.create_attribute("test_attr", float, 1, dense=True)
    try:
        attr[4] = 42.
        assert False
    except M.mesh.mesh_attributes.Attribute.OutOfBoundsError:
        assert True

    try:
        attr[-1] = 42.
        assert False
    except M.mesh.mesh_attributes.Attribute.OutOfBoundsError:
        assert True

def test_attribute_types_dense(test_container):
    attr_bool    = test_container.create_attribute("test_bool",    bool,    1, dense=True)
    attr_int     = test_container.create_attribute("test_int",     int,     1, dense=True)
    attr_float   = test_container.create_attribute("test_float",   float,   1, dense=True)
    attr_complex = test_container.create_attribute("test_complex", complex, 1, dense=True)
    attr_str     = test_container.create_attribute("test_str",     str,     1, dense=True)

    attr_bool[0] = True
    assert attr_bool[0] == True

    attr_int[0] = 42
    assert attr_int[0] == 42
    
    attr_float[0] = 23.52
    assert attr_float[0] == 23.52

    attr_complex[0] = 10+22.5j
    assert attr_complex[0] == 10+22.5j

    attr_str[0] = 'bonjour'
    assert attr_str[0] == 'bonjour'


def test_attribute_types_vec_dense(test_container):
    attr_bool    = test_container.create_attribute("test_bool",    bool,    2, dense=True)
    attr_int     = test_container.create_attribute("test_int",     int,     2, dense=True)
    attr_float   = test_container.create_attribute("test_float",   float,   2, dense=True)
    attr_complex = test_container.create_attribute("test_complex", complex, 2, dense=True)
    attr_str     = test_container.create_attribute("test_str",     str,     2, dense=True)

    attr_bool[0] = [True, False]
    assert (attr_bool[0] == [True, False]).all()

    attr_int[0] = [42, 10]
    assert (attr_int[0] == [42,10]).all()
    
    attr_float[0] = [23.52, 35.12]
    assert (attr_float[0] == [23.52, 35.12]).all()

    attr_float[0][0] = 11.111
    assert (attr_float[0] == [11.111, 35.12]).all()

    attr_complex[0] = [10+22.5j, 9+9j]
    assert (attr_complex[0] == [10+22.5j, 9+9j] ).all()

    attr_str[0] = ["bonjour", "merci"]
    assert (attr_str[0] == ["bonjour", "merci"]).all()



from ...mesh.datatypes import *
from ...mesh.mesh import copy
from ...mesh.mesh_attributes import Attribute, ArrayAttribute
from ...geometry import Vec
from ..worker import Worker
from ... import geometry as geom
from ...attributes.glob import euler_characteristic

import numpy as np
from math import atan2
from scipy.sparse.linalg import lsqr, eigsh
from scipy.sparse import coo_matrix

class LSCM(Worker):
    """
    Least-Square Conformal Map algorithm for computing a parametrization of a mesh.
    /!\\ The mesh should have the topology of a disk.

    Usage :
    ```
    lscm = LSCM(mesh)
    lscm.run([options])
    ```

    or

    ```
    lscm = LSCM(mesh)([options]) # directly calls run
    ```

    This fills the self.uvs container
    """

    @allowed_mesh_types(SurfaceMesh)
    def __init__(self, mesh : SurfaceMesh, verbose:bool=True):
        super().__init__("LSCM", verbose=verbose)
        self.mesh : SurfaceMesh = mesh
        self._flat_mesh : SurfaceMesh = None
        self.uvs : Attribute = None # attribute on vertices 

    def run(self, eigen=True, save_on_mesh = True, axis_align = True, verbose=True):
        self.verbose = verbose
        if euler_characteristic(self.mesh)!=1:
            raise Exception("Mesh is not a topological disk. Cannot run LSCM.")
        if eigen:
            U = self._solve_eigen()
        else:
            U = self._solve()
        U = U.reshape((U.size//2, 2))
        if axis_align:
            U = self._align_with_axes(U)
        U = self._scale(U)
        if save_on_mesh:
            self.uvs = self.mesh.vertices.create_attribute("uv_coords", float, 2, dense=True)
        else:
            self.uvs = ArrayAttribute(float, len(self.mesh.vertices), 2)
        for v in self.mesh.id_vertices:
            self.uvs[v] = Vec(U[v])

    def _build_system(self):
        n = len(self.mesh.vertices)
        m = len(self.mesh.faces)
        cols, rows, vals = (np.zeros(12*m) for _ in range(3))
        for t,T in enumerate(self.mesh.faces):
            i,j,k = T
            Pi, Pj, Pk = (self.mesh.vertices[_v] for _v in (i,j,k))
            X,Y,_ = geom.face_basis(Pi,Pj,Pk)
            Qi, Qj, Qk = ( Vec(X.dot(_P), Y.dot(_P)) for _P in (Pi,Pj,Pk))
            
            for _i, (r,c,v) in enumerate([
                # first row
                (2*t, 2*i, Qk[0] - Qj[0]),
                (2*t, 2*j, Qi[0] - Qk[0]),
                (2*t, 2*k, Qj[0] - Qi[0]),
                (2*t, 2*i+1, Qk[1] - Qj[1]),
                (2*t, 2*j+1, Qi[1] - Qk[1]),
                (2*t, 2*k+1, Qj[1] - Qi[1]),
                # second row
                (2*t+1, 2*i, Qk[1] - Qj[1]),
                (2*t+1, 2*j, Qi[1] - Qk[1]),
                (2*t+1, 2*k, Qj[1] - Qi[1]),
                (2*t+1, 2*i+1, Qj[0] - Qk[0]),
                (2*t+1, 2*j+1, Qk[0] - Qi[0]),
                (2*t+1, 2*k+1, Qi[0] - Qj[0])
            ]):
                rows[12*t+_i], cols[12*t+_i], vals[12*t+_i] = r, c, v 
        A = coo_matrix((vals, (rows, cols)),  shape=(2*m,2*n)).tolil()
        self.log("Size of system :", A.shape)
        return A

    def _solve(self):
        self.log("Building system...")
        n = len(self.mesh.vertices)
        A = self._build_system()

        # index locked 1 has value (0,0) and index locked2 has value (1,1)
        locked1 = np.random.randint(0,n)
        locked2 = np.random.randint(0,n)
        while locked2 == locked1:
            locked2 = np.random.randint(0,n)
        locked1, locked2 = min(locked1, locked2), max(locked1, locked2)

        B = np.squeeze(np.array(A.getcol(2*locked2).todense() + A.getcol(2*locked2+1).todense()))
        inds = [x for x in range(A.shape[1]) if x not in {2*locked1, 2*locked1+1, 2*locked2, 2*locked2+1}]
        A = A[:, inds] # remove corresponding columns 
        self.log("Solving system...")
        res = lsqr(A, -B)
        U_int = res[0]
        self.log("Done. Residual norm ||AX-B||=",res[3])
        U = np.zeros((2*n,))
        U[:2*locked1] = U_int[:2*locked1]
        U[2*locked1+2:2*locked2] = U_int[2*locked1:2*locked2-2]
        U[2*locked2] = 1
        U[2*locked2 + 1] = 1
        U[2*locked2+2:] = U_int[2*locked2-2:]
        return U

    def _solve_eigen(self):
        self.log("Building system...")
        A = self._build_system()
        B = A.transpose() @ A
        self.log("Solving for eigenvector...")
        egv, U = eigsh(B, 3, which="SM", tol=1e-4)
        self.log("Done. Smallest eigenvalues:", egv)
        U = U[:,0]
        return U

    def _scale(self, U):
        # scale UVs in bounding box [0;1]^2
        xmin,xmax,ymin,ymax = float("inf"), -float("inf"), float("inf"), -float("inf")
        for i in range(U.shape[0]):
            u,v = U[i][0], U[i][1]
            xmin = min(xmin, u)
            xmax = max(xmax, u)
            ymin = min(ymin, v)
            ymax = max(ymax, v)
        scale_x = xmax-xmin
        scale_y = ymax-ymin
        scale = min(scale_x, scale_y)
        # apply scale
        return U/scale

    def _align_with_axes(self,U):
        # rotate I.UVs so that feature edges are axis aligned
        orig = U[0]
        e = self.mesh.boundary_edges[0] # should exist since mesh is disk topology
        A,B = self.mesh.edges[e]
        vec = U[A] - U[B]
        angle = -atan2(vec[1], vec[0])
        # apply transformation
        for v in self.mesh.id_vertices:
            uvi = Vec(U[v]-orig)
            U[v] = geom.rotate_2d(uvi, angle) + orig
        return U

    @property
    def flat_mesh(self):
        if self.uvs is None:
            return None
        if self._flat_mesh is None:
            # build the flat mesh : vertex coordinates are uv of original mesh
            self._flat_mesh = copy(self.mesh)
            for i in self.mesh.id_vertices:
                self._flat_mesh.vertices[i] = Vec(self.uvs[i][0], self.uvs[i][1], 0.)
        return self._flat_mesh


"""
An integer (2^32 -1) representing a default value in geogram file format
"""
NOT_AN_ID = 4294967295

"""
In most file formats, surface meshes are given as a set of vertices and a set of faces, since edges can be retrieved from faces.
If this option is set to true, the edge set of the mesh will be constructed from the face information.
"""
complete_edges_from_faces = True

"""
In most file formats, volume meshes are given as a set of vertices and tetrahedra (cells).
If this option is set to true, the face set of the mesh will be constructed from the cells information.
"""
complete_faces_from_cells = True


"""
Whether to export edges as 'l <v> <v>' fields in .obj file format
"""
export_edges_in_obj = True

"""
/!\ If set to true, this will compute the half edge structure on the mesh (which can cause a lot of computations)
"""
sort_neighborhoods = True

"""
When creating an attribute that already exists on a mesh, mouette prints a warning in the console and returns the attribute currently carrying this name.
If this flag is set to True, no warning is printed. We then assume that you know what you are doing
"""
disable_duplicate_attribute_warning = True


import numpy as np

class Vec(np.ndarray):
    
    ###### Constructors ######
    def __new__(cls, *a):
        if len(a)==1:
            obj = np.asarray(a[0]).view(cls)
        else:
            obj = np.asarray(a).view(cls)
        return obj
    
    @staticmethod
    def random(n : int):
        data = np.random.random(n)
        return Vec(data)

    @staticmethod
    def zeros(n : int):
        data = np.zeros(n)
        return Vec(data)
    
    ###### Accessors ######
    @property
    def x(self):
        return self[0]

    @x.setter
    def x(self, value):
        self[0] = value

    @property
    def y(self):
        return self[1]

    @y.setter
    def y(self, value):
        self[1] = value

    @property
    def xy(self):
        return self[:2]

    @property
    def z(self):
        return self[2]

    @z.setter
    def z(self, value):
        self[2] = value

    ###### Methods ######
    def norm(self, which="l2") -> float :
        if which=="l2":
            return np.sqrt(np.dot(self, self))
        elif which=="l1":
            return np.sum(np.abs(self))
        elif which=="linf":
            return np.max(np.abs(self))

    def dot(self, other) -> float:
        return np.dot(self, other)

    def outer(self,other):
        return np.outer(self,other)

    def normalize(self, which="l2"):
        self /= self.norm(which)

    @staticmethod
    def normalized(vec, which="l2"):
        nrm = Vec.norm(vec, which)
        np.seterr(all='raise')
        # we want the following to fail when a division by zero is encountered
        out = Vec(vec/nrm)
        np.seterr(all='warn')
        return out


import numpy as np
from math import pi, cos, sin
from scipy.spatial import ConvexHull

from ..geometry import Vec,cross,dot
from ..mesh.mesh_data import RawMeshData
from ..mesh.datatypes import *
from ..mesh.mesh import merge, _instanciate_raw_mesh_data

from ..geometry import transform

from .platonic import icosahedron

def sphere_uv( n_lat : int, n_long : int, center : Vec = Vec(0.,0.,0.), radius : float = 1.) -> SurfaceMesh:
    """Generates a surface mesh using uv sampling of a sphere

    Args:
        n_lat (int): number of different latitudes for points
        n_long (int): number of different longitudes for points
        center (Vec, optional): Center position of the sphere. Defaults to Vec(0.,0.,0.).
        radius (float, optional): Radius of the sphere. Defaults to 1.

    Returns:
        SurfaceMesh: the sphere
    """
    sp = RawMeshData()
    # add two points at poles
    sp.vertices.append(center + radius*Vec(0.,0.,1.))
    sp.vertices.append(center + radius*Vec(0.,0.,-1.))

    # add other points
    theta = np.linspace(-pi/2,pi/2, n_long+2)[1:-1]
    phi = np.linspace(-pi, pi,n_lat+1)[:-1]
    for t in theta:
        for p in phi:
            sp.vertices.append(Vec(radius*cos(t)*cos(p), radius*cos(t)*sin(p), radius*sin(t)))

    # build surface as convex hull
    ch = ConvexHull(sp.vertices._data, qhull_options="QJ")
    
    # correct normals
    for face in ch.simplices:
        pA,pB,pC = (sp.vertices[_u] for _u in face)
        ray = (pA+pB+pC)/3 - center
        nrml = cross(pB-pA,pC-pA)
        if dot(ray,nrml)<0:
            sp.faces.append([face[0], face[2], face[1]])
        else:
            sp.faces.append(face)
    return _instanciate_raw_mesh_data(sp,2)

def icosphere(n_refine : int= 3, center : Vec = Vec(0.,0.,0.), radius : float = 1.) -> SurfaceMesh :
    ico = icosahedron(center)
    ico = transform.scale(ico, radius, center)
    # Subdivide
    for _ in range(n_refine):
        raise NotImplementedError
    return ico

def spherify_vertices(points : PointCloud, radius : float = 1e-2, n_subdiv=1) -> SurfaceMesh:
    spheres = []
    if isinstance(points, Mesh):
        what = points.vertices
    else:
        what = points
    for P in what:
        mp = icosphere(n_subdiv, P, radius)
        spheres.append(mp)
    return merge(spheres)

def sphere_fibonacci( n_pts : int , build_surface : bool = True) -> SurfaceMesh:
    """Generates a point cloud or a surface mesh using fibonacci sampling of a sphere.

    Args:
        n_pts (int): total number of vertices
        build_surface (bool, optional): If specified to True, the function will also compute a triangulation of the vertices. This is obtained through a convex hull algorithm (since points lay on a convex shape, the convex hull and the Delaunay triangulation are equivalent). Defaults to True.

    Returns:
        [SurfaceMesh | PointCloud]: the generated mesh
    """

    phi = 0.5 * (1. + np.sqrt(5.))

    theta = np.zeros(n_pts)
    sphi = np.zeros(n_pts)
    cphi = np.zeros(n_pts)

    points = []
    for i in range(n_pts):
        j = 2*i - (n_pts-1) 
        theta = 2.0 * np.pi *  j / phi
        sphi = j / n_pts
        cphi = np.sqrt( (n_pts + j ) * ( n_pts - j )) /  n_pts

        x = cphi * np.sin(theta)
        y = cphi * np.cos(theta)
        z = sphi
        points.append(Vec(x,y,z))

    data = RawMeshData()
    data.vertices += points
    if build_surface:
        # Triangulate points on a sphere : Delaunay is equivalent to convex hull since every points lay in the hull
        ch = ConvexHull(points, qhull_options="QJ")
        data.faces += list(ch.simplices)
    return _instanciate_raw_mesh_data(data)



import numpy as np
from scipy.linalg import expm
from scipy.spatial.transform import Rotation
from math import cos, sin, sqrt, tan, atan
from .vector import Vec
from .geometry import norm, dot
from .rotations import axis_rot_from_z

"""
Angular momentum operators for spherical harmonics representation.
Three matrices correspond to three axes.
# taken from Algebraic Representations for Volumetric Frame Fields, DAVID PALMER, DAVID BOMMES and JUSTIN SOLOMON (Supplementary materials)
"""
LX = np.array([
    [0.,      0.,        0.,        0.,        0.,        0.,         0.,         -sqrt(2),   0.      ],
    [0.,      0.,        0.,        0.,        0.,        0.,         -sqrt(7/2), 0.,         -sqrt(2)],
    [0.,      0.,        0.,        0.,        0.,        -3/sqrt(2), 0.,         -sqrt(7/2), 0.      ],
    [0.,      0.,        0.,        0.,        -sqrt(10), 0.,         -3/sqrt(2), 0.,         0.      ],
    [0.,      0.,        0.,        sqrt(10),  0.,        0.,         0.,         0.,         0.      ],
    [0.,      0.,        3/sqrt(2), 0.,        0.,        0.,         0.,         0.,         0.      ],
    [0.,      sqrt(7/2), 0.,        3/sqrt(2), 0.,        0.,         0.,         0.,         0.      ],
    [sqrt(2), 0.,        sqrt(7/2), 0.,        0.,        0.,         0.,         0.,         0.      ],
    [0.,      sqrt(2),   0.,        0.,        0.,        0.,         0.,         0.,         0.      ],
])

LY = np.array([
    [0.,       sqrt(2),    0.,         0.,        0.,       0.,        0.,         0.,         0.      ],
    [-sqrt(2), 0.,         sqrt(7/2),  0.,        0.,       0.,        0.,         0.,         0.      ],
    [0.,       -sqrt(7/2), 0.,         3/sqrt(2), 0.,       0.,        0.,         0.,         0.      ],
    [0.,       0.,         -3/sqrt(2), 0.,        0.,       0.,        0.,         0.,         0.      ],
    [0.,       0.,         0.,         0.,        0.,       -sqrt(10), 0.,         0.,         0.      ],
    [0.,       0.,         0.,         0.,        sqrt(10), 0.,        -3/sqrt(2), 0.,         0.      ],
    [0.,       0.,         0.,         0.,        0.,       3/sqrt(2), 0.,         -sqrt(7/2), 0.      ],
    [0.,       0.,         0.,         0.,        0.,       0.,        sqrt(7/2),  0.,         -sqrt(2)],
    [0.,       0.,         0.,         0.,        0.,       0.,        0.,         sqrt(2),    0.      ],
])

LZ = np.array([
    [0.,  0.,  0.,  0., 0., 0., 0., 0., 4.],
    [0.,  0.,  0.,  0., 0., 0., 0., 3., 0.],
    [0.,  0.,  0.,  0., 0., 0., 2., 0., 0.],
    [0.,  0.,  0.,  0., 0., 1., 0., 0., 0.],
    [0.,  0.,  0.,  0., 0., 0., 0., 0., 0.],
    [0.,  0.,  0., -1., 0., 0., 0., 0., 0.],
    [0.,  0., -2.,  0., 0., 0., 0., 0., 0.],
    [0., -3.,  0.,  0., 0., 0., 0., 0., 0.],
    [-4., 0.,  0.,  0., 0., 0., 0., 0., 0.],
])

# hard coded rotation matrices that are often used
Rxpi4 = expm(np.pi/4 * LX)
Rypi4 = expm(np.pi/4 * LY)
Rzpi4 = expm(np.pi/4 * LZ)
Rxzpi4 = Rzpi4 @ Rxpi4
Rxpi2 = expm(np.pi/2 * LX)
Rypi2 = expm(np.pi/2 * LY)

def RZ(a : float):
    c = [cos(k*a) for k in range(5)]
    s = [sin(k*a) for k in range(5)]
    return np.array([
        [ c[4], 0.,   0.,   0.,   0., 0.,   0.,   0.,   s[4]],
        [ 0.,   c[3], 0.,   0.,   0., 0.,   0.,   s[3], 0.  ],
        [ 0.,   0.,   c[2], 0.,   0., 0.,   s[2], 0.,   0.  ],
        [ 0.,   0.,   0.,   c[1], 0., s[1], 0.,   0.,   0.  ],
        [ 0.,   0.,   0.,   0.,   1., 0.,   0.,   0.,   0.  ],
        [ 0.,   0.,   0.,  -s[1], 0., c[1], 0.,   0.,   0.  ],
        [ 0.,   0.,  -s[2], 0.,   0., 0.,   c[2], 0.,   0.  ],
        [ 0.,  -s[3], 0.,   0.,   0., 0.,   0.,   c[3], 0.  ],
        [-s[4], 0.,   0.,   0.,   0., 0.,   0.,   0.,   c[4]],
    ])

def RY(a : float):
    return Rxpi2.T @ RZ(a) @ Rxpi2

def RX(a : float):
    return Rypi2 @ RZ(a) @ Rypi2.T

# coefficients of the identity matrix
EYE = np.array([0., 0., 0., 0., sqrt(7/12), 0., 0., 0., sqrt(5/12) ])

def skew_matrix_from_rotvec(w : Vec) -> np.ndarray :
    return w[0] * LX + w[1] * LY + w[2] * LZ

def rot_matrix_from_euler(w : Vec) -> np.ndarray :
    return RX(w[0]) @ RY(w[1]) @ RZ(w[2])

def rot_matrix_from_rotvec(v : Vec) -> np.ndarray :
    return expm(skew_matrix_from_rotvec(v))
    #euler = Rotation.from_rotvec(v).as_euler("XYZ")
    #return rot_matrix_from_euler(euler)

def from_vec3(v : Vec) -> np.ndarray:
    if norm(v)<1e-8: return np.array([0.,0.,0.,0.,1.,0.,0.,0.,0.])
    R = rot_matrix_from_rotvec(v)
    sh = R[:,4] # apply rotation to representation of (0,0,1) vector : 1 at 4th coeff and 0 otherwise -> dot product is fourth column of matrix
    return sh

def from_frame(frame : Rotation) -> np.ndarray:
    axis = Vec(frame.as_euler("XYZ"))
    if norm(axis)<1e-8: return EYE
    R = rot_matrix_from_euler(axis)
    sh = R.dot(EYE) # apply found rotation to the spherical harmonic reference
    return sh

def rotate_frame(sh : np.ndarray, r : Rotation) -> np.ndarray :
    if isinstance(r, Rotation):
        return Vec(rot_matrix_from_euler(r.as_rotvec()).dot(sh))
    return Vec(rot_matrix_from_euler(r).dot(sh))

def project_to_frame(sh : Vec, stop_threshold : float = 1e-8, max_iter=100, nrml_cstr: Vec = None):
    """Given the coefficients in the spherical harmonic basis, finds the frames that correspond the most.

        Also recomputes spherical harmonics coefficients to a perfect match.
        Uses the Cayleigh transform to approximate the 9D rotation.

    Args:
        sh (Vec): the 9 coefficients representing the frame in spherical harmonics basis
        stop_threshold (float,optional) : Stoping criterion on the norm of the residual. Defaults to 1e-8.
        max_iter (int, optional): Maximum number of gradient steps. Defaults to 100.
        nrml_cstr (Vec) : a 3 dimensionnal normal direction, to constraint the projection to only be around an axis. Defaults to None.
    
    Returns:
        frame (scipy.spatial.transform.Rotation) : the obtained frame
        a (Vec) : the projected and updated spherical harmonics coefficients
    """
    if norm(sh)<1e-6: return Rotation.identity(), sh
    q = Vec.normalized(sh)
    
    if nrml_cstr is None:
        seeds = [EYE] + [R.dot(EYE) for R in (Rxpi4, Rypi4, Rzpi4, Rxzpi4)]
        pi4 = np.pi/4
        frames = [Rotation.identity(),
                Rotation.from_euler("x", pi4),
                Rotation.from_euler("y", pi4),
                Rotation.from_euler("z", pi4),
                Rotation.from_euler("xz", (pi4,pi4))]
        dots = [ np.dot(q, seed) for seed in seeds]
        seed_choice = np.argmax(dots)
        a = seeds[seed_choice]
        frame = frames[seed_choice]
        for _ in range(max_iter):
            s, b = (q+a)/2, (q-a)
            l = np.array([ LX.dot(s), LY.dot(s), LZ.dot(s)])
            if norm( l.dot(b))<stop_threshold: break
            x = np.linalg.lstsq(l.T, b, rcond=None)[0]
            x = Vec.normalized(x) * 2*atan(norm(x)/2)
            rot = Rotation.from_rotvec(x)
            frame = rot * frame
            x = rot.as_euler("XYZ")
            a = rot_matrix_from_euler(x).dot(a)
    else:
        ax = axis_rot_from_z(nrml_cstr)
        frames = [ R * Rotation.from_rotvec(ax) for R in 
                    (Rotation.identity(),
                     Rotation.from_rotvec(np.pi/4 * nrml_cstr))] 
        seeds = [from_frame(_frame) for _frame in frames]
        dots = [ np.dot(q, seed) for seed in seeds]
        seed_choice = np.argmax(dots)
        a = seeds[seed_choice]
        frame = frames[seed_choice]
        L = skew_matrix_from_rotvec(nrml_cstr)
        for _ in range(max_iter):
            s,b = L.dot(q+a), q-a
            if abs(s.dot(b))<stop_threshold: break
            alpha = dot(s,b)/dot(s,s)
            rot = Rotation.from_rotvec(2*atan(alpha)*nrml_cstr)
            frame = rot * frame
            x = rot.as_euler("XYZ")
            a = rot_matrix_from_euler(x).dot(a)
    return frame,a

def project_to_frame_grad(sh : Vec, lr : float = 1e-1, grad_threshold : float = 1e-4, dot_threshold = 1e-6, max_iter=1000):
    """Given the coefficients in the spherical harmonic basis, finds the frames that correspond the most. Computation is unfortunately not direct:
        starting from the reference frame, we perform a gradient descent on the l2 distance between spherical harmonics coefficients.

        Also recomputes spherical harmonics coefficients to a perfect match.
        Uses the linearization of the exponential to approximate 9D rotations.

        /!\\ This algorithm is less precise and less efficient than project_to_frame. Use the latter instead.

    Args:
        sh (Vec): the 9 coefficients representing the frame in spherical harmonics basis
        lr (float, optional): Gradient descent learning rate. Defaults to 1.
        grad_threshold (float,optional) : Stoping criterion on the norm of the gradient. Defaults to 1e-4
        dot_threshold (float,optional) : Stoping criterion on the norm of the step. Defaults to 1e-6
        max_iter (int, optional): Maximum number of gradient steps. Defaults to 1000.
    
    Returns:
        frame (scipy.spatial.transform.Rotation) : the obtained frame
        a (Vec) : the projected and updated spherical harmonics coefficients
    """

    if norm(sh)<1e-6: return Rotation.identity(), sh
    Id = np.eye(9)
    q = Vec.normalized(sh)
    Rx, Ry, Rz = (Id + np.pi/4 * L for L in (LX, LY, LZ))
    qLX, qLY, qLZ = (np.dot(L.T, q) for L in (LX, LY, LZ))
    seeds = [EYE] + [R.dot(EYE) for R in (Rxpi4, Rypi4, Rzpi4, Rxzpi4)]
    pi4 = np.pi/4
    frames = [Rotation.identity(),
              Rotation.from_euler("x", pi4),
              Rotation.from_euler("y", pi4),
              Rotation.from_euler("z", pi4),
              Rotation.from_euler("xz", (pi4,pi4))]
    dots = [ np.dot(q, seed) for seed in seeds]
    seed_choice = np.argmax(dots)
    old_dot = dots[seed_choice]

    a = seeds[seed_choice]
    frame = frames[seed_choice]
    n_iter = 0
    while n_iter<max_iter:
        gx,gy,gz = ( np.dot(_q, a) for _q in (qLX,qLY,qLZ))
        grad = Vec( gx, gy, gz ) # gradient in point a
        if grad.norm()< grad_threshold: break

        step = lr*grad
        # technically, Rx, Ry and Rz should be exponentials of L1, L2, L3 with corresponding angles,
        # but we approximate at first order exp(X) = 1 + X + o(|X|)
        Rx = Id + step.x * LX
        Ry = Id + step.y * LY
        Rz = Id + step.z * LZ
        
        R = Rx @ Ry @ Rz
        a = R.dot(a)
        
        new_dot = np.dot(q,a)
        if abs(old_dot-new_dot)<dot_threshold: break
        old_dot = new_dot

        frame = Rotation.from_euler("XYZ", [step.x, step.y, step.z]) * frame
        n_iter += 1
    return frame,a

def orthogonality_energy(sh : np.ndarray) -> float:
    """Computes an energy that is 0 if and only if the given spherical harmonics has octahedral symetry (ie represents an orthogonal frame)
    The energy is quadratic in terms of the harmonic coefficients

    Args:
        sh (np.ndarray): a spherical harmonics (9 coefficients)

    Returns:
        (float): the computed energy
    """
    s1 = 28*(sh[0]*sh[0] + sh[8]*sh[8]) + 7*(sh[1]*sh[1] + sh[7]*sh[7]) - 8*(sh[2]*sh[2] + sh[6]*sh[6]) - 17*(sh[3]*sh[3] + sh[5]*sh[5]) - 20*sh[4]*sh[4]
    s1 *= sqrt(2)

    s2 = 28*(sh[0]*sh[1] + sh[7]*sh[8]) + 10*sqrt(7)*(sh[1]*sh[2] + sh[6]*sh[7]) + 18*(sh[2]*sh[3] + sh[5]*sh[6]) + 4*sqrt(5)*sh[4]*sh[5]
    s2 *= sqrt(3)

    s3 = 28*(sh[0]*sh[7] - sh[1]*sh[8]) + 10*sqrt(7)*(sh[1]*sh[6] - sh[2]*sh[7]) + 4*sqrt(5)*sh[4]*sh[3] + 18*(sh[2]*sh[5] - sh[3]*sh[6])
    s3 *= sqrt(3)

    s4 = 4*sqrt(7)*(sh[0]*sh[2] + sh[6]*sh[8]) + 6*sqrt(7)*(sh[1]*sh[3] + sh[5]*sh[7]) + 12*sqrt(5)*sh[4]*sh[6] + 10*(sh[3]*sh[3] - sh[5]*sh[5])
    s4 *= sqrt(6)

    s5 = 4*sqrt(7)*(sh[6]*sh[0] - sh[2]*sh[8]) + 6*sqrt(7)*(sh[1]*sh[5] - sh[3]*sh[7]) + 12*sqrt(5)*sh[2]*sh[4] - 20*sh[3]*sh[5]
    s5 *= sqrt(6)

    e = np.array((s1,s2,s3,s4,s5))
    return np.dot(e,e)


from ..mesh_data import RawMeshData
from ..mesh_attributes import Attribute
from .base import Mesh
from .linear import PolyLine
from ... import utils
from ... import config

class SurfaceMesh(Mesh):

    def __init__(self, data : RawMeshData = None):
        Mesh.__init__(self, 2, data)

        if self.is_triangular():
            self.half_edges = SurfaceMesh._HalfEdgeStructure(self)
        else:
            self.half_edges = None

        self.connectivity = SurfaceMesh._Connectivity(self)

        # Boundary data
        self._boundary_edges : list = None
        self._interior_edges : list = None

        self._is_vertex_on_border : Attribute = None 
        self._boundary_vertices : list = None
        self._interior_vertices : list = None

    def __str__(self):
        out = "mouette.mesh.SurfaceMesh object\n"
        out += "| {} vertices\n".format(len(self.vertices))
        out += "| {} edges\n".format(len(self.edges))
        out += "| {} faces\n".format(len(self.faces))
        return out

    @property
    def id_vertices(self):
        """
        Shortcut for range(len(self.vertices))
        """
        return range(len(self.vertices))

    @property
    def id_edges(self):
        """
        Shortcut for range(len(self.edges))
        """
        return range(len(self.edges))

    @property
    def id_faces(self):
        """
        Shortcut for range(len(self.faces))
        """
        return range(len(self.faces))

    @property
    def id_corners(self):
        """
        Shortcut for range(len(self.face_corners))
        """
        return range(len(self.face_corners))

    def ith_vertex_of_face(self, fid, vid):
        return self.faces[fid][vid]

    def pt_of_face(self,fid):
        return (self.vertices[_v] for _v in self.faces[fid])

    def is_triangular(self) -> bool:
        """
        Returns:
            bool: True if the mesh is triangular (all faces are triangles)
        """
        for f in self.faces:
            if len(f)!=3: return False
        return True

    def is_quad(self) -> bool:
        for f in self.faces:
            if len(f) != 4: return False
        return True

    def recompute_edges(self):
        """Recomputes the set of edges according to the set of faces """
        self.edges.clear()
        edge_set = set()
        for f in self.faces:
            nf = len(f)
            for i in range(nf):
                edge = utils.keyify(f[i], f[(i+1)%nf])
                if edge not in edge_set:
                    edge_set.add(edge)
                    self.edges.append(edge)

    def clear_boundary_data(self):
        self._boundary_edges : list = None
        self._interior_edges : list = None

        self._is_vertex_on_border : Attribute = None 
        self._boundary_vertices : list = None
        self._interior_vertices : list = None

    ##### Accessors for boundary and exterior ######

    def _compute_interior_boundary_edges(self):
        self._interior_edges = []
        self._boundary_edges = []
        for (u,v) in self.half_edges:
            id = self.connectivity.edge_id(u,v)
            if self.is_edge_on_border(u,v):
                self._boundary_edges.append(id)
            else:
                self._interior_edges.append(id)
    
    def _compute_interior_boundary_vertices(self):
        self._boundary_vertices = set()
        self._is_vertex_on_border = self.vertices.create_attribute("border", bool)
        for e in self.boundary_edges:
            a,b = self.edges[e]
            self._is_vertex_on_border[a] = True
            self._is_vertex_on_border[b] = True
            self._boundary_vertices.add(a)
            self._boundary_vertices.add(b)
        self._boundary_vertices = list(self.boundary_vertices)
        
        self._interior_vertices = []
        for x in self.id_vertices:
            if not self._is_vertex_on_border[x]:
                self._interior_vertices.append(x)

    def is_edge_on_border(self,u,v) -> bool:
        return self.half_edges.adj(u,v)[0] is None or self.half_edges.adj(v,u)[0] is None

    def is_vertex_on_border(self,u) -> bool:
        if self._is_vertex_on_border is None:
            self._compute_interior_boundary_vertices()
        return self._is_vertex_on_border[u]

    @property
    def interior_edges(self):
        if self._interior_edges is None:
            self._compute_interior_boundary_edges()
        return self._interior_edges

    @property
    def boundary_edges(self):
        if self._boundary_edges is None:
            self._compute_interior_boundary_edges()
        return self._boundary_edges

    @property
    def boundary_vertices(self):
        if self._boundary_vertices is None:
            self._compute_interior_boundary_vertices()
        return self._boundary_vertices
    
    @property
    def interior_vertices(self):
        if self._interior_vertices is None:
            self._compute_interior_boundary_vertices()
        return self._interior_vertices

    ###### Half Edges ######

    class _HalfEdgeStructure:
        def __init__(self, master):
            self.mesh : SurfaceMesh = master
            self._half_edges : dict = None

        def clear(self):
            self._half_edges : dict = None

        def _compute_half_edges(self):
            self._half_edges = dict()
            for i,F in enumerate(self.mesh.faces):
                n = len(F)
                for v in range(n):
                    self._half_edges[ (F[v], F[(v+1)%n] ) ] = (i,v,(v+1)%n)

        def __iter__(self):
            if self._half_edges is None:
                self._compute_half_edges()
            return self._half_edges.__iter__()

        def adj(self, u : int, v : int):
            """Pair (u,v) of vertex -> triangle to the left of edge (u,v) if edge (u,v) exists, None otherwise
            Also returns local indexes of u and v in the triangle (and None if (u,v) does not exists)
            
            Calling this function with edge (v,u) yield the triangle of the other side of the edge
            """
            if self._half_edges is None:
                self._compute_half_edges()
            if (u,v) in self._half_edges:
                return self._half_edges[(u,v)]
            return None,None,None

        def opposite(self, u : int, v : int, T : int):
            """Given a pair of vertices (u,v) and a face T, returns the face (and local indexes of u and v) on the other side of edge (u,v)

            if (u,v) are not two vertices of the face T, returns None
            """
            T1,u1,v1 = self.adj(u,v)
            T2,v2,u2 = self.adj(v,u)
            if T1==T:
                return T2,u2,v2
            if T2==T:
                return T1,u1,v1
            return None,None,None

        def edge_to_triangles(self, u, v):
            return self.adj(u,v)[0], self.adj(v,u)[0]

        def common_edge(self,iF1,iF2):
            F1 = self.mesh.faces[iF1]
            n = len(F1)
            for i in range(n):
                A,B = F1[i], F1[(i+1)%n]
                if self.opposite(A,B,iF1)[0]==iF2:
                    return A,B
            return None,None

        def next_around(self, v : int, u: int):
            """ 
            Turning clockwise in the neighbourhood of vertex v, finds the vertex after u.
            Returns None is such a vertex does not exist (reach boundary)
            """
            T, iV, iU = self.adj(v,u)
            if T is None: return None
            return self.mesh.ith_vertex_of_face(T, 3-iV-iU)

        def prev_around(self, v, u):
            """ 
            Turning counter clockwise in the neighbourhood of vertex v, finds the vertex after u.
            Returns None is such a vertex does not exist (reach boundary)
            """
            T, iU, iV = self.adj(u,v)
            if T is None: return None
            return self.mesh.ith_vertex_of_face(T, 3-iV-iU)

    ###### connectivity ######

    class _Connectivity(PolyLine._Connectivity):

        def __init__(self, master):
            super().__init__(master)

            self._adjV2F : dict = dict() # vertex -> triangle
            # V2F is not None because connectivity is built on the fly at each query
            
            self._adjF2F : dict = None # triangle -> triangle
            self._adjVF2Cn : dict = None # vertex,face -> corner
            self._adjCn2F : dict = None # corner -> face
            self._adjF2Cn : dict = None # face -> first corner

            self._face_id : dict = None
        
        def clear(self):
            super().clear()
            self._adjV2F : dict = dict() # vertex -> triangle            
            self._adjF2F : dict = None # triangle -> triangle
            self._adjVF2Cn : dict = None # vertex,face -> corner
            self._adjCn2F : dict = None # corner -> face
            self._adjF2Cn : dict = None # face -> first corner
            self._face_id : dict = None     

        def face_id(self, a, b, c):
            if self._face_id is None:
                self._compute_vertex_adj()
            key = utils.keyify(a,b,c)
            return self._face_id.get(key, None)

        def _compute_vertex_adj(self):
            super()._compute_vertex_adj()

            self._adjV2F = dict([(i,set()) for i in self.mesh.id_vertices])
            self._face_id = dict()

            for iF,F in enumerate(self.mesh.faces):
                key = utils.keyify(F)
                self._face_id[key] = iF
                for V in F:
                    self._adjV2F[V].add(iF)

            for U in self.mesh.id_vertices:
                self._adjV2F[U] = list(self._adjV2F[U])
            
            # Neighborhood should be sorted to have a clockwise order
            if config.sort_neighborhoods and isinstance(self.mesh, SurfaceMesh):
                self._sort_vertex_neighborhoods()

        def _sort_vertex_neighborhoods(self):
            for A in self.mesh.interior_vertices:
                if not self._adjV2V[A] : continue # no adjacent vertices to sort
                indexV, indexF = dict(), dict()
                B = self._adjV2V[A][0]
                T = self.mesh.half_edges.adj(A,B)[0]
                indexV[B] = 0
                indexF[T] = 0
                for i in range(len(self._adjV2V[A])):
                    B = self.mesh.half_edges.next_around(A,B) # B before T since root triangle is handled outside of the loop
                    T = self.mesh.half_edges.adj(A,B)[0]
                    indexV[B] = i+1
                    indexF[T] = i+1

                self._adjV2V[A].sort(key = lambda u : indexV[u])
                self._adjV2F[A].sort(key = lambda u : indexF[u])

            for A in self.mesh.boundary_vertices:
                if not self._adjV2V[A] : continue # not adjacent vertices to sort
                indexV,indexF = dict(), dict()
                B = self._adjV2V[A][0]
                ind = 0
                indexV[B] = ind
                while self.mesh.half_edges.adj(B,A)[0] is not None:
                    T = self.mesh.half_edges.adj(B,A)[0] # T before B because root triangle is not handled outside of the loop
                    B = self.mesh.half_edges.prev_around(A,B)
                    ind-=1
                    indexV[B] = ind
                    indexF[T] = ind
                # reset and go the other way around
                B = self._adjV2V[A][0]
                ind = 0
                while self.mesh.half_edges.adj(A,B)[0] is not None:
                    T = self.mesh.half_edges.adj(A,B)[0] # T before B because root triangle is not handled outside of the loop
                    B = self.mesh.half_edges.next_around(A,B)
                    ind +=1
                    indexV[B] = ind
                    indexF[T] = ind
                    
                self._adjV2V[A].sort(key = lambda u : indexV[u])
                self._adjV2F[A].sort(key = lambda u : indexF[u]) 
        
        def _compute_face_adj(self):
            self._adjF2F = dict([(t, []) for t in self.mesh.id_faces])
            for (A,B) in self.mesh.edges:
                F1, _, _ = self.mesh.half_edges.adj(A,B)
                F2, _, _ = self.mesh.half_edges.adj(B,A)
                if F1 is not None and F2 is not None:
                    self._adjF2F[F1].append(F2)
                    self._adjF2F[F2].append(F1)

        def _compute_corner_adj(self):
            self._adjVF2Cn = dict()
            self._adjCn2F = dict()
            self._adjF2Cn = dict()
            c = 0
            for iF,F in enumerate(self.mesh.faces):
                self._adjF2Cn[iF] = c
                for v in F:
                    assert self.mesh.face_corners[c] == v
                    self._adjVF2Cn[(v,iF)] = c
                    self._adjCn2F[c] = iF
                    c += 1

        ##### Vertex to Faces #####

        def n_VtoF(self, V):
            L = self._adjV2F.get(V, None)
            if L is None:
                self._adjV2F[V] = []
                for iT,T in enumerate(self.mesh.faces):
                    if V in T:
                        self._adjV2F[V].append(iT)
            return len(self._adjV2F[V])

        def vertex_to_face(self, V):
            L = self._adjV2F.get(V, None)
            if L is None :
                self._compute_vertex_adj()
            return self._adjV2F[V]
        
        ##### Corners #####

        def vertex_to_corner_in_face(self, V, F):
            if self._adjVF2Cn is None:
                self._compute_corner_adj()
            return self._adjVF2Cn.get((V,F), None)

        def corner_to_face(self,C):
            if self._adjCn2F is None:
                self._compute_corner_adj()
            return self._adjCn2F.get(C,None)
        
        def face_to_first_corner(self,F):
            if self._adjF2Cn is None:
                self._compute_corner_adj()
            return self._adjF2Cn[F]

        def face_to_corners(self,F):
            if self._adjF2Cn is None:
                self._compute_corner_adj()
            return [self._adjF2Cn[F] + _i for _i in range(len(self.mesh.faces[F]))]

        ##### Faces to Vertex #####

        def face_to_vertex(self, F):
            return list(self.mesh.faces[F])

        def in_face_index(self, F, V):
            """Index of vertex V in face F. None if V is not in face F

            Args:
                F (int): face index
                V (int): vertex index

            Returns:
                int
            """
            for (i,v) in enumerate(self.mesh.faces[F]):
                if v==V: return i
            return None
        
        ##### Face to Edge ######

        def face_to_edge(self, F):
            lF = self.mesh.faces[F]
            n = len(lF)
            return [self.edge_id(lF[i],lF[(i+1)%n]) for i in range(n)]

        ###### Face to face ######

        def face_to_face(self, F):
            if self._adjF2F is None:
                self._compute_face_adj()
            return self._adjF2F[F]


import numpy as np

from ..geometry import Vec
from .platonic import hexahedron
from ..mesh.datatypes import *

def axis_aligned_cube(colored=False) -> SurfaceMesh:
    """generated an axis aligned cube as 6 quad faces.

       7--------6
      /|       /|
     / |      / |
    4--------5  |
    |  |     |  |
    |  3-----|--2
    | /      | /
    |/       |/
    0--------1

    Args:
        colored (bool, optional): if set to true, will add a colo rattribute on faces to determine. Defaults to False.

    Returns:
        SurfaceMesh: a cube
    """
    v0 = Vec(-0.5,-0.5,-0.5)
    v1 = Vec(0.5,-0.5,-0.5)
    v2 = Vec(0.5,0.5,-0.5)
    v3 = Vec(-0.5,0.5,-0.5)

    v4 = Vec(-0.5,-0.5,0.5)
    v5 = Vec(0.5,-0.5,0.5)
    v6 = Vec(0.5,0.5,0.5)
    v7 = Vec(-0.5,0.5,0.5)
    return hexahedron(v0,v1,v2,v3,v4,v5,v6,v7, colored)

def hexahedron_4pts(P1 : Vec, P2 : Vec, P3 : Vec, P4 : Vec, colored=False, volume=False) -> SurfaceMesh:
    """Generate an hexahedron given by an absolute position and three points building a basis.

    4
    |
    |  3
    | /
    |/
    1--------2

    Args:
        P1 to P4 (Vec): coordinates of vertices
        colored (bool, optional): if set to true, will add a color attribute on faces to determine. Defaults to False.
        volume (bool, optional): if set to true, will also generate three tetrahedra to fill the volume. Defaults to False.

    Returns:
        SurfaceMesh: [description]
    """
    X,Y = P2-P1, P3-P1
    return hexahedron(P1, P1+X, P1+X+Y, P1+Y, P4, P4+X, P4+X+Y, P4+Y, colored, volume)


import mouette as M
from data import *
import pytest

@pytest.mark.parametrize("m", surfaces)
def test_feature_detect(m):
    feat = M.processing.FeatureEdgeDetector(only_border= False, flag_corners=True)
    feat.run(m)
    assert isinstance(feat.feature_vertices, set)
    assert isinstance(feat.feature_graph, M.mesh.PolyLine)


import mouette as M
from utils import *
from data import *

## Vertices
@pytest.mark.parametrize("s", surfaces + polylines)
def test_degree(s):
    deg = M.attributes.degree(s)
    deg = M.attributes.degree(s, persistent=False)
    deg = M.attributes.degree(s, dense=False)
    assert True

@pytest.mark.parametrize("s", surfaces)
def test_angle_defect(s):
    dfct = M.attributes.angle_defects(s)
    dfct = M.attributes.angle_defects(s, persistent=False)
    dfct = M.attributes.angle_defects(s, dense=False)
    assert True

@pytest.mark.parametrize("s", surfaces)
def test_vertex_normals(s):
    vn = M.attributes.vertex_normals(s)
    vn = M.attributes.vertex_normals(s, persistent=False)
    vn = M.attributes.vertex_normals(s, dense=False)
    assert True

@pytest.mark.parametrize("s", surfaces)
def test_border_normals(s):
    bn = M.attributes.border_normals(s) # not dense by default
    bn = M.attributes.border_normals(s, persistent=False)
    bn = M.attributes.border_normals(s, dense=True)
    assert True


import mouette as M

def test_vec_zeros_constructor():
    V = M.Vec.zeros(5)
    assert V.size == 5
    for i in range(5):
        assert V[i] == 0

def test_vec_random_constructor():
    V = M.Vec.random(10)
    assert V.size == 10

def test_vec_from_list():
    lst = [x for x in range(10)]
    V = M.Vec(lst)
    assert V.size==10

def test_vec_xyz_accessor():
    V = M.Vec([1,0,0])
    assert V.x == 1
    assert V.y == 0
    assert V.z == 0

def test_vec_xyz_setters():
    V = M.Vec.zeros(4)
    V.x = 42
    V.y = 64
    V.z = 12
    assert V.x == 42
    assert V[0] == 42
    assert V.y == 64
    assert V[1] == 64
    assert V.z == 12
    assert V[2] == 12
    assert V[3] == 0

def test_vec_bracket_accessor():
    lst = [x for x in range(10)]
    V = M.Vec(lst)
    for i in range(10):
        assert V[i] == i


from .mesh_attributes import Attribute, ArrayAttribute
from ..geometry import Vec
from .. import utils
from .. import config
import itertools
import warnings
from .. import config

class DataContainer:
    """
    A DataContainer is a container class for all simplicial elements of the same type in an instance (for example, vertices, edges or faces).
    It stores relevant information about the combinatorics (in the _data field) as well as various attributes onto the elements (in the _attr field)
    """

    def __init__(self, data : list = None, attributes : dict = None, id : str = ""):
        self.id = id
        if data is None:
            self._data = []
        else:
            self._data = list(data)
        
        # self._attr : dict of str->Attribute
        if attributes is None:
            self._attr = dict()
        else:
            assert isinstance(attributes, dict)
            self._attr = attributes 
    
    def __getitem__(self, key):
        return self._data[key]

    def __setitem__(self, key, value):
        try:
            self._data[key] = value
        except Exception as e:
            warnings.warn(f"Error in attribute '{key}' : {e}")
            raise Exception("Aborting")

    def __call__(self, name):
        if name not in self._attr:
            raise Exception("Attribute does not exist")
        return self._attr[name]

    def __iter__(self):
        return self._data.__iter__()

    def __repr__(self):
        return self._data.__repr__()

    def __str__(self):
        return str(self._data) + "\nAttributes: "+str(self._attr.keys())

    @property
    def size(self):
        return len(self._data)

    def __len__(self):
        return len(self._data)

    def empty(self):
        return not self._data

    def clear(self):
        self._data = []
        self._attr = dict()

    @property
    def attributes(self):
        return self._attr.keys()

    def create_attribute(self, name, data_type, elem_size=1, dense=False, default_value=None) -> Attribute:
        # If 'name' already exists in the attribute dict, the corresponding attribute will be overridden
        if name in self._attr and not config.disable_duplicate_attribute_warning:
            warnings.warn(f"Warning ! Attribute '{name}' already exists on {self.id}")
        else:
            if dense:
                self._attr[name] = ArrayAttribute(data_type, len(self), elem_size=elem_size, default_value=default_value)
            else:
                self._attr[name] = Attribute(data_type, elem_size=elem_size, default_value=default_value)
        return self._attr[name]

    def delete_attribute(self, name):
        if name in self._attr:
            del self._attr[name]

    def has_attribute(self, name) -> bool:
        return name in self._attr

    def get_attribute(self, name) -> Attribute :
        if name not in self._attr:
            raise Exception("Attribute does not exist")
        return self._attr[name]

    def append(self,val):
        self._data.append(val)
        for attr in self._attr.values():
            attr._expand(1)

    def __iadd__(self,other):
        if isinstance(other, list) or isinstance(other,tuple) or isinstance(other, set):
            self._data += list(other)
            for attr in self._attr.values():
                attr._expand(len(other))
        elif isinstance(other,DataContainer):
            self._data += other._data
            for attr in self._attr.values():
                attr._expand(other.n_elem)
        else:
            raise Exception("Could not append data container of type {} onto an attribute".format(type(other)))
        return self


class RawMeshData:
    """
    A base container class to store all the data relative to a mesh.
    It is not made to be used outside of the library, and serves an intermediate between io parsers and instantiated (and typed) mesh classes.
    """

    def __init__(self, mesh=None):
        self.vertices = DataContainer(id="vertices") if mesh is None else mesh.vertices
        self.edges = DataContainer(id="edges") if (mesh is None or not hasattr(mesh,"edges")) else mesh.edges

        self.faces = DataContainer(id="faces") if (mesh is None or not hasattr(mesh,"faces")) else mesh.faces
        self.face_corners = DataContainer(id="face_corners") if (mesh is None or not hasattr(mesh,"face_corners")) else mesh.face_corners
       
        self.cells = DataContainer(id="cells") if (mesh is None or not hasattr(mesh,"cells")) else mesh.cells
        self.cell_corners = DataContainer(id="cell_corners") if (mesh is None or not hasattr(mesh,"cell_corners")) else mesh.cell_corners
        self.cell_faces = DataContainer(id="cell_faces") if (mesh is None or not hasattr(mesh,"cell_faces")) else mesh.cell_faces

        self._dimensionality : int = None
        self._prepared : bool = False
    
    @property
    def id_vertices(self):
        """
        Shortcut for range(len(self.vertices))
        """
        return range(len(self.vertices))

    @property
    def id_edges(self):
        """
        Shortcut for range(len(self.edges))
        """
        return range(len(self.edges))

    @property
    def id_faces(self):
        """
        Shortcut for range(len(self.faces))
        """
        return range(len(self.faces))

    @property
    def id_cells(self):
        """
        Shortcut for range(len(self.cells))
        """
        return range(len(self.cells))

    @property
    def dimensionality(self) -> int:
        """
        Dimensionality of the manifold represented by the data.
        If only vertices -> 0
        If vertices and edges (embedded graph) -> 1
        If faces (surface manifold) -> 2
        If cells (volume manifold) -> 3

        Returns:
            int
        """
        if self._dimensionality is None:
            self._compute_dimensionality()
        return self._dimensionality

    def _compute_dimensionality(self):
        if not self.cells.empty():
            self._dimensionality = 3
        elif not self.faces.empty():
            self._dimensionality = 2
        elif not self.edges.empty():
            self._dimensionality = 1
        else:
            self._dimensionality = 0

    def prepare(self) -> None:
        """
        Prepares the data to have the correct format.
        On vertices : casts to M.Vec
        On edges : sorts edge tuples to satisfy edge convention (smallest index first)
        On faces : nothing
        On cells : nothing

        if option is set in config, adds to the edge container all the edges implicitly defined by the faces
        """
        if self._prepared : return

        # call completion before preparation
        if config.complete_faces_from_cells:
            self._complete_faces_from_cells()

        # call edge completion ***after*** face completion
        if config.complete_edges_from_faces : 
            self._complete_edges_from_faces() 

        self._prepare_vertices()
        self._prepare_edges()
        self._prepare_faces()
        self._prepare_face_corners()
        self._prepare_cells()
        self._compute_dimensionality()
        self._prepared = True

    def _prepare_vertices(self):
        for iv in self.id_vertices:
            self.vertices[iv] = Vec(self.vertices[iv])

    def _prepare_edges(self):
        N = len(self.vertices)

        def is_valid(a,b):
            return a!=b and 0<=a<N and 0<=b<N

        # Filter invalid edges (like (x,x)) and make the other immutable
        edges_invalid = any((not is_valid(a,b) for a,b in self.edges))
        if edges_invalid:
            # Rebuild the edge container
            new_edges = DataContainer(id="edges")
            new_attrs = dict()
            old_attrs = dict()
            for attr_name in self.edges.attributes:
                old_attrs[attr_name] = self.edges.get_attribute(attr_name)
                new_attrs[attr_name] = new_edges.create_attribute(attr_name, old_attrs[attr_name].type, old_attrs[attr_name].elemsize)
            n = 0
            for ie in self.id_edges:
                a,b = self.edges[ie]
                if is_valid(a,b):
                    new_edges.append(utils.keyify(a,b))
                    for name in new_attrs:
                        if ie in old_attrs[name]: # keep sparsity of the attribute
                            new_attrs[name][n] = old_attrs[name][ie]
                    n+=1
            self.edges = new_edges
        else:
            for ie in self.id_edges:
                self.edges[ie] = utils.keyify(self.edges[ie])

    def _prepare_faces(self):
        pass
        # for iF in self.id_faces:
        #     self.faces[iF] = tuple(self.faces[iF])
        

    def _prepare_face_corners(self):
        nc = len(self.face_corners)
        nf = sum([len(f) for f in self.faces])
        if nc == 0 or nc!= nf : 
            # corners were not generated
            self.face_corners._data = []
            for f in self.faces:
                self.face_corners._data += list(f)

    def _prepare_cells(self):
        pass

    def _complete_edges_from_faces(self):
        if self.faces.empty() : return # nothing to do
        
        # Create hard edges attribute for already existing edges
        hard_edges = self.edges.create_attribute("hard_edges", bool)
        for e in self.id_edges: hard_edges[e] = True
        
        edge_set = set([utils.keyify(e) for e in self.edges])
        for f in self.faces:
            nf = len(f)
            for i in range(nf):
                edge = utils.keyify(f[i], f[(i+1)%nf])
                if edge not in edge_set:
                    edge_set.add(edge)
                    self.edges.append(edge)

    def _complete_faces_from_cells(self):
        if self.cells.empty() : return # nothing to do
        face_set = set([utils.keyify(f) for f in self.faces])
        for C in self.cells:
            if len(C)>4:
                continue
            for F in itertools.combinations(C,3): # all possibilities of 3-uples among the tetrahedron
                face = utils.keyify(F)
                if face not in face_set:
                    face_set.add(face)
                    self.faces.append(F)


import mouette as M
from data import *

@pytest.mark.parametrize("m", surfaces)
def test_cotan_laplacian(m):
    mat = M.operators.laplacian(m, False)
    n = len(m.vertices)
    assert mat.shape == (n,n)
    
@pytest.mark.parametrize("m", surfaces + polylines)
def test_graph_laplacian(m):
    mat = M.operators.graph_laplacian(m)
    n = len(m.vertices)
    assert mat.shape == (n,n)

@pytest.mark.parametrize("m", surfaces)
def test_weight_matrix(m):
    mat = M.operators.area_weight_matrix(m)
    n = len(m.vertices)
    assert mat.shape == (n,n)

@pytest.mark.parametrize("m", surfaces + polylines)
def test_adjacency_matrix(m):
    mat = M.operators.adjacency_matrix(m)
    n = len(m.vertices)
    assert mat.shape == (n,n)


from .faces2d import FrameField2DFaces, CurvatureFaces
from .vertex2d import FrameField2DVertices, CadFF2DVertices, CurvatureVertices
from .cells import FrameField3DCells
from .vertex3d import FrameField3DVertices


from ..mesh_data import RawMeshData
from .base import Mesh

class PointCloud(Mesh):

    def __init__(self, data : RawMeshData = None):
        Mesh.__init__(self, 0, data)

    def __str__(self):
        out = "mouette.mesh.PointCloud object\n"
        out += "| {} vertices\n".format(len(self.vertices))
        return out

    def append(self,x):
        # shortcut since we can only append elements in the 'vertices' container
        self.vertices.append(x)

    @property
    def id_vertices(self):
        """
        Shortcut for range(len(self.vertices))
        """
        return range(len(self.vertices))


import mouette as M
from data.volumes import *
from utils import *

def test_new():
    m = M.mesh.new_volume()
    assert len(m.vertices)==0
    assert len(m.edges)==0
    assert len(m.faces)==0
    assert len(m.cells)==0

### Io tests ###

# @pytest.mark.parametrize("v", volumes)
# def test_io_medit(v, tmp_path):
#     build_test_io(v, tmp_path, "mesh", 3)

# @pytest.mark.parametrize("v", volumes)
# def test_io_geogram_ascii(v, tmp_path):
#     build_test_io(v, tmp_path, "geogram_ascii", 3)

# @pytest.mark.parametrize("v", volumes)
# def test_io_tet(v, tmp_path):
#     build_test_io(v, tmp_path, "tet", 3)


from .base import FrameField
from ...mesh.datatypes import *
from ...mesh.mesh_attributes import Attribute, ArrayAttribute
from ... import geometry as geom
from ...geometry import Vec
from ... import operators
from ...utils.maths import roots, angle_diff
from ...attributes import cotangent, angle_defects, mean_edge_length, curvature_matrices
from ..features import FeatureEdgeDetector

import numpy as np
from math import pi, atan2
import cmath
import scipy.sparse as sp
from scipy.sparse import linalg

class _BaseFrameField2DFaces(FrameField) : 

    @allowed_mesh_types(SurfaceMesh)
    def __init__(self, supporting_mesh : SurfaceMesh, order:int = 4, feature_edges : bool = False, verbose:bool=True):
        super().__init__(verbose=verbose)
        self.mesh : SurfaceMesh = supporting_mesh
        self.order : int = order

        self.cot : Attribute = None
        self.defect : Attribute = None
        self.tbaseX : Attribute = None # local basis X vector (on triangles)
        self.tbaseY : Attribute = None # local basis Y vector (on triangles)
        self.tnormals : Attribute = None # local basis Z vector (= mesh normal at triangles)

        self.features : bool = feature_edges # whether feature edges are enabled or not
        self.feat : FeatureEdgeDetector = None
        
        self.initialized = False

    def _initialize_attributes(self):
        #processing.split_double_boundary_edges_triangles(self.mesh) # A triangle has only one edge on the boundary
        self.cot = cotangent(self.mesh)
        self.defect = angle_defects(self.mesh,persistent=False)

    def _initialize_features(self):
        self.feat = FeatureEdgeDetector(only_border = not self.features, verbose=self.verbose)(self.mesh)

    def _initialize_bases(self):
        NF = len(self.mesh.faces)
        self.tbaseX, self.tbaseY = ArrayAttribute(float, NF, 3), ArrayAttribute(float, NF, 3)
        self.tnormals = self.mesh.faces.create_attribute("normals", float, 3, dense=True)
        for id_face, (A,B,C) in enumerate(self.mesh.faces):
            pA,pB,pC = (self.mesh.vertices[_v] for _v in (A,B,C))
            X,Y,Z = geom.face_basis(pA,pB,pC) # local basis of the triangle
            self.tbaseX[id_face] = X
            self.tbaseY[id_face] = Y
            self.tnormals[id_face] = Z

    def _initialize_variables(self):
        self.var = np.zeros(len(self.mesh.faces), dtype=complex)

        # fix orientation on features
        for e in self.feat.feature_edges:
            e1,e2 = self.mesh.edges[e] # the edge on border
            edge = self.mesh.vertices[e2] - self.mesh.vertices[e1]
            for T in self.mesh.half_edges.edge_to_triangles(e1,e2):
                if T is None: continue # edge may be on boundary
                X,Y = self.tbaseX[T], self.tbaseY[T]
                c = complex(edge.dot(X), edge.dot(Y)) # compute edge in local basis coordinates (edge.dot(Z) = 0 -> complex number for 2D vector)
                self.var[T] = (c/abs(c))**4 # c^4 is the same for all four directions of the cross

    def flag_singularities(self):
        """Compute singularity data.

        Creates 2 attributes:
            - an attribute "singuls" on vertices storing the value (+- 1) of singularities (eventually 0 for non singular vertices)
            - an attribute "angles" on edges storing the angle of the edge, given as the difference between the two frames
        """
        self._check_init()
        ZERO_THRESHOLD = 1e-3

        if self.mesh.edges.has_attribute("angles"):
            edge_rot = self.mesh.edges.get_attribute("angles")
            edge_rot.clear()
        else:
            edge_rot = self.mesh.edges.create_attribute("angles", float, 1, dense=True)
            # the rotation induced by the frame field on every edge
            # if edge is uv, positive orientation is from T(uv) to T(vu)

        for ie,(A,B) in enumerate(self.mesh.edges):
            T1,T2 = self.mesh.half_edges.edge_to_triangles(A,B)
            if T1 is None or T2 is None: continue
            f1,f2 = self.var[T1], self.var[T2] # representation complex for T1 and T2
            
            # parallel transport
            E = geom.Vec.normalized(self.mesh.vertices[B] - self.mesh.vertices[A])
            a1 = atan2(self.tbaseY[T1].dot(E), self.tbaseX[T1].dot(E))
            a2 = atan2(self.tbaseY[T2].dot(E), self.tbaseX[T2].dot(E))

            # matching
            u2 = roots(f2, self.order)[0]
            angles = [angle_diff( cmath.phase(u2) - a2, cmath.phase(u1) - a1) for u1 in roots(f1, self.order)]
            abs_angles = [abs(_a) for _a in angles]
            i_angle = np.argmin(abs_angles)
            edge_rot[ie] = angles[i_angle]

        if self.mesh.vertices.has_attribute("singuls"):
            singuls = self.mesh.vertices.get_attribute("singuls")
            singuls.clear()
        else:
            singuls = self.mesh.vertices.create_attribute("singuls", float)
        
        for v in self.mesh.id_vertices:
            angle = self.defect[v]
            for e in self.mesh.connectivity.vertex_to_edge(v):
                u = self.mesh.connectivity.other_edge_end(e,v)
                angle += edge_rot[e] if u<v else -edge_rot[e]
            if abs(angle)>ZERO_THRESHOLD:
                singuls[v] = angle*2/pi

    def export_as_mesh(self) -> PolyLine:
        FFMesh = PolyLine()
        L = mean_edge_length(self.mesh)/3
        for id_face, face in enumerate(self.mesh.faces):
            pA,pB,pC = (self.mesh.vertices[u] for u in face)
            basis,_,normal = geom.face_basis(pA,pB,pC)

            angle = cmath.phase(self.var[id_face])/4
            bary = (pA+pB+pC)/3 # reference point for display
            r1,r2,r3,r4 = (geom.rotate_around_axis(basis, normal, angle + k*pi/2) for k in range(4))
            p1,p2,p3,p4 = (bary + abs(self.var[id_face])*L*r for r in (r1,r2,r3,r4))
            FFMesh.vertices += [bary, p1, p2, p3, p4]
            FFMesh.edges += [(5*id_face, 5*id_face+k) for k in range(1,5)]            
        return FFMesh

class FrameField2DFaces(_BaseFrameField2DFaces) : 

    @allowed_mesh_types(SurfaceMesh)
    def __init__(self, supporting_mesh : SurfaceMesh, order:int = 4,  feature_edges : bool = False, verbose:bool=True):
        super().__init__(supporting_mesh, order,feature_edges, verbose)
           
    def initialize(self):
        self._initialize_attributes() # /!\ may change mesh combinatorics near boundary
        self._initialize_features()
        self._initialize_bases()
        self._initialize_variables()
        self.initialized = True

    def optimize(self, n_renorm=10):
        self._check_init()
    
        self.log("Build laplacian operator")
        lap = operators.laplacian_triangles(self.mesh, cotan=False, order=self.order).tocsc()
        A = operators.area_weight_matrix_faces(self.mesh).tocsc().astype(complex)

        ###### Border ######
        if len(self.feat.feature_vertices)>0: # We have a border / feature elements -> linear solve
            # Build fixed and variable indexes
            self.log("Feature element detected (border and/or feature edges)")

            fixed = self.mesh.faces.create_attribute("fixed", bool)
            for ie in self.feat.feature_edges:
                u,v = self.mesh.edges[ie]
                T1,T2 = self.mesh.half_edges.edge_to_triangles(u,v)
                if T1 is not None: fixed[T1] = True
                if T2 is not None: fixed[T2] = True
            freeInds,fixedInds = [],[]
            for T in self.mesh.id_faces:
                if fixed[T] : fixedInds.append(T)
                else: freeInds.append(T)
              
            if len(freeInds)==0 : 
                self.log("Everything is on boundary : no optimization required")
                return

            lapI = lap[freeInds,:][:,freeInds]
            lapB = lap[freeInds,:][:,fixedInds]
            # for lapI and lapB, only lines of freeInds are relevant : lines of fixedInds link fixed variables -> useless constraints
            AI = A[freeInds, :][:, freeInds]
            valB = lapB.dot(self.var[fixedInds]) # right hand side
            
            self.log("Initial solve of linear system")
            res = linalg.spsolve(lapI, -valB) # first system solved without diffusion
            self.var[freeInds] = res

            if n_renorm>0:
                self.log(f"Solve linear system {n_renorm} times with diffusion")
                alpha = 1e-3 * mean_edge_length(self.mesh)
                self.log("Attach weight: {}".format(alpha))
                mat = lapI - alpha * AI
                for _ in range(n_renorm):
                    self.normalize()
                    valI2 = alpha * AI.dot(self.var[freeInds])
                    res = linalg.spsolve(mat, -valB - valI2)
                    self.var[freeInds] = res
            self.normalize()

        ###### No border ######
        else:
            self.log("No border detected")
            self.log("Initial solve of linear system using an eigensolver")
            try:
                egv, U = sp.linalg.eigsh(lap, k=1, M=A, which="SM", tol=1e-4,maxiter=1000)
            except linalg.ArpackNoConvergence as e:
                self.log("Initial eigensolve failed :", e)
                self.log("Retry using connectivity laplacian")
                lap = operators.laplacian_triangles(self.mesh, cotan=False).tocsc()
                egv, U = sp.linalg.eigsh(lap, k=2, M=A, which="SM", tol=1e-4)

            alpha = abs(egv[0])
            self.log("Eigenvalues:", egv)
            self.var = U[:,0]

            if n_renorm>0:
                self.log(f"Solve linear system {n_renorm} times with diffusion")
                mat = lap - alpha * A
                for _ in range(n_renorm):
                    self.normalize()
                    valI2 = alpha * A.dot(self.var)
                    self.var = linalg.spsolve(mat, -valI2)
            self.normalize()

class CurvatureFaces(_BaseFrameField2DFaces):

    def __init__(self, supporting_mesh : SurfaceMesh, verbose :bool = False):
        super().__init__(supporting_mesh, 4, False, verbose)
        self.curv_mat_edges : np.ndarray = None # shape (|E|,3,3)
        self.curv_mat_faces : np.ndarray = None # shape (|T|,3,3)

    def _initialize_curvature(self):
        self.curv_mat_edges = curvature_matrices(self.mesh)
        adj_e = dict([(e,set()) for e in self.mesh.id_edges])
        for T in self.mesh.id_faces:
            for T2 in self.mesh.connectivity.face_to_face(T):
                for e in self.mesh.connectivity.face_to_edge(T2):
                    adj_e[e].add(T)

        self.curv_mat_faces = np.zeros((len(self.mesh.faces),3,3))
        for e in self.mesh.id_edges:
            for T in adj_e[e]:
                self.curv_mat_faces[T] += self.curv_mat_edges[e]

    def run(self, n_smooth=0):
        self.initialize()
        self.log("Optimize")
        self.optimize(n_smooth)
        self.log("Done.")
        return self

    def initialize(self):
        self.log("Initialize attributes")
        self._initialize_attributes() # /!\ may change mesh combinatorics near boundary
        self._initialize_features()
        self.log("Initialize bases")
        self._initialize_bases()
        self.log("Compute curvature matrices")
        self._initialize_curvature()
        self.log("Initialize variables")
        self._initialize_variables()
        self.initialized = True

    def optimize(self, n_smooth):
        for T in self.mesh.id_faces:
            if abs(self.var[T])!=0: continue # already initialized => T is on the boundary
            u,s,_ = np.linalg.svd(self.curv_mat_faces[T,:,:], hermitian=True)
            X,Y = self.tbaseX[T], self.tbaseY[T]
            # three vectors -> normal and two principal components
            # eigenvalue of normal is 0
            # PC are orthogonal (eigenvects of symmetric matrix) -> we rely on the eigenvect of greatest eigenvalue and take orthog direction
            if s[0]<1e-10 : # zero matrix -> no curvature information -> we take the first basis vector and hope to smooth correctly
                self.var[T] = 0 + 0j
            else:
                # representation vector
                v = u[:,0]
                c = complex(X.dot(v), Y.dot(v))
                self.var[T] = (c/abs(c))**4

        if n_smooth > 0:
            # diffuse the curvature results to get a smoother results (especially where curvature was not defined)
            lap = operators.laplacian_triangles(self.mesh, order=self.order)
            A = operators.area_weight_matrix_faces(self.mesh).astype(complex)
            alpha = 1e-3

            if len(self.feat.feature_vertices)>0: # we have a boundary
                pass
            
            else:
                mat = lap - alpha * A
                for _ in range(n_smooth):
                        valI2 = alpha * A.dot(self.var)
                        self.var = linalg.spsolve(mat, - valI2)
                        self.normalize()


import numpy as np

from ..mesh.mesh_attributes import ArrayAttribute, Attribute
from ..mesh.datatypes import *
from .. import geometry as geom
from ..geometry import Vec

@allowed_mesh_types(SurfaceMesh, VolumeMesh)
def face_area(mesh : Mesh, name="area", persistent:bool=True, dense:bool=True):
    """
    Computes the barycenter point of each face.

    Args:
        mesh (SurfaceMesh)
        name (str, optional): Name given to the attribute. Defaults to "area".
        persistent (bool, optional): If the attribute is persistent (stored in the mesh object) or not. Defaults to True.
        dense (bool, optional): Is the attribute dense (numpy array) or not (dict). Defaults to True

    Returns:
        Attribute[float] on faces
    """
    if persistent:
        area = mesh.faces.create_attribute(name, float, dense=dense)
    else:
        area = ArrayAttribute(float, len(mesh.faces)) if dense else Attribute(float)
    for T in mesh.id_faces:
        pts = [mesh.vertices[u] for u in mesh.faces[T]]
        npt = len(pts)
        if npt==3:
            area[T] = geom.triangle_area(*pts)
        elif npt==4:
            area[T] = geom.quad_area(*pts)
        else:
            bary = sum(pts)/npt
            for i in range(npt):
                A = pts[i]
                B = pts[(i+1)%npt]
                area[T] += geom.triangle_area(A,B,bary)            
    return area

@allowed_mesh_types(SurfaceMesh)
def face_normals(mesh : SurfaceMesh, name="normals", persistent:bool=True, dense:bool=True):
    """
    Computes the barycenter point of each face.

    Args:
        mesh (SurfaceMesh)
        name (str, optional): Name given to the attribute. Defaults to "normals".
        persistent (bool, optional): If the attribute is persistent (stored in the mesh object) or not. Defaults to True.
        dense (bool, optional): Is the attribute dense (numpy array) or not (dict). Defaults to True

    Returns:
        Attribute[float, 3] on faces
    """
    if persistent :
        normals = mesh.faces.create_attribute(name, float, 3)
    else:
        normals = ArrayAttribute(float, len(mesh.faces), 3) if dense else Attribute(float)
    for iT,T in enumerate(mesh.faces):
        pA,pB,pC = (mesh.vertices[u] for u in T[:3])
        normals[iT] = Vec.normalized(geom.cross(pB-pA, pC-pA))
    return normals

@allowed_mesh_types(SurfaceMesh, VolumeMesh)
def face_barycenter(mesh : SurfaceMesh, name="barycenter", persistent:bool = True, dense:bool = True):
    """
    Computes the barycenter point of each face.

    Args:
        mesh (SurfaceMesh)
        name (str, optional): Name given to the attribute. Defaults to "barycenter".
        persistent (bool, optional): If the attribute is persistent (stored in the mesh object) or not. Defaults to True.
        dense (bool, optional): Is the attribute dense (numpy array) or not (dict). Defaults to True

    Returns:
        Attribute[float, 3] on faces
    """
    if persistent :
        bary = mesh.faces.create_attribute(name, float, 3)
    else:
        bary = ArrayAttribute(float, len(mesh.faces), 3) if dense else Attribute(float, 3)
    for iT,T in enumerate(mesh.faces):
        bary[iT] = sum(mesh.vertices[u] for u in T) / len(T)
    return bary

@allowed_mesh_types(SurfaceMesh, VolumeMesh)
def face_circumcenter(mesh : SurfaceMesh, name="circumcenter", persistent:bool=True, dense:bool=True):
    """
    Computes the circumcenter point of each face.

    Args:
        mesh (SurfaceMesh)
        name (str, optional): Name given to the attribute. Defaults to "barycenter".
        persistent (bool, optional): If the attribute is persistent (stored in the mesh object) or not. Defaults to True.
        dense (bool, optional): Is the attribute dense (numpy array) or not (dict). Defaults to True

    Returns:
        Attribute[float, 3] on faces
    """
    if persistent :
        circum = mesh.faces.create_attribute(name, float, 3)
    else:
        circum = ArrayAttribute(float, len(mesh.faces), 3) if dense else Attribute(float, 3)
    for iF,F in enumerate(mesh.faces):
        if len(F)!=3:
            raise Exception("circumcenters can be computed only for triangular faces. Received face n°{} containing {}!=3 vertices".format(iF,len(F)))
        circum[iF] = geom.circumcenter(*(mesh.vertices[u] for u in F))

@allowed_mesh_types(SurfaceMesh)
def faces_near_border(mesh : SurfaceMesh, dist:int = 2, name = "near_border", persistent:bool = True, dense:bool = False):
    """Returns the faces that are at most at 'dist' neighbours from the boundary.
    Proceeds by region growing, starting from all faces touching the boundary and going inwards.

    Args:
        mesh (SurfaceMesh): input mesh
        dist (int, optional): Extend to which we flag the faces. All faces with a path of length < dist to the boundary will be returned. Defaults to 2.
        name (str, optional): Name given to the attribute. Defaults to "barycenter".
        persistent (bool, optional): If the attribute is persistent (stored in the mesh object) or not. Defaults to True.
        dense (bool, optional): Is the attribute dense (numpy array) or not (dict). Defaults to False

    Returns:
        Attribute[bool] on faces. If not persistent, returns a set.
    """
    
    near = set()
    for e in mesh.boundary_edges:
        T1,T2 = mesh.half_edges.edge_to_triangles(*mesh.edges[e])
        if T1 is None: near.add(T2)
        else: near.add(T1)
    for _ in range(dist-1):
        new_near = set()
        for T in mesh.id_faces:
            if T in near : continue
            for T2 in mesh.connectivity.face_to_face(T):
                if T2 in near: new_near.add(T)
        near = near | new_near
    
    if persistent:
        near_attr = mesh.faces.create_attribute(name, bool, dense=dense)
    else:
        near_attr = ArrayAttribute(bool, len(mesh.faces)) if dense else Attribute(bool)
    for f in near: near_attr[f] = True
    return near_attr



from mouette.processing import framefield as GEOFF
from data import *

@pytest.mark.parametrize("m", [surf_circle(), surf_spline(), surf_half_sphere(), surf_pointy()])
def test_FF2DVertices(m):
    ff = GEOFF.FrameField2DVertices(m, 4, False, False)()
    ff.flag_singularities()
    assert m.faces.has_attribute("singuls")

@pytest.mark.parametrize("m", [surf_circle(), surf_spline(), surf_half_sphere(), surf_pointy()])
def test_CadFF2DVertices(m):
    ff = GEOFF.CadFF2DVertices(m, 4, False)()
    ff.flag_singularities()
    assert m.faces.has_attribute("singuls")

@pytest.mark.parametrize("m", [surf_circle(), surf_spline(), surf_half_sphere(), surf_pointy()])
def test_CurvatureFFVertices(m):
    ff = GEOFF.CurvatureVertices(m, False)()
    ff.flag_singularities()
    assert m.faces.has_attribute("singuls")